"""Routines and classes for storing groups important to PROPKA calculations."""
import math
import propka.ligand
import propka.protonate
from propka.ligand_pka_values import LigandPkaValues
from propka.determinant import Determinant
from propka.lib import info, warning


# Constants that start with "UNK_" are a mystery to me
UNK_PKA_SCALING = -1.36
PROTONATOR = propka.protonate.Protonate(verbose=False)
EXPECTED_ATOMS_ACID_INTERACTIONS = {
    'COO': {'O': 2}, 'HIS': {'H': 2, 'N': 2}, 'CYS': {'S': 1}, 'TYR': {'O': 1},
    'LYS': {'N': 1}, 'ARG': {'H': 5, 'N': 3}, 'ROH': {'O': 1},
    'AMD': {'H': 2, 'N': 1}, 'TRP': {'H': 1, 'N': 1}, 'N+': {'N': 1},
    'C-': {'O': 2}, 'BBN': {'H': 1, 'N': 1,}, 'BBC': {'O': 1},
    'NAR': {'H': 1, 'N': 1}, 'NAM': {'H': 1, 'N': 1}, 'F': {'F': 1},
    'Cl': {'Cl': 1}, 'OH': {'H': 1, 'O': 1}, 'OP': {'O': 1}, 'O3': {'O': 1},
    'O2': {'O': 1}, 'SH': {'S': 1}, 'CG': {'H': 5, 'N': 3},
    'C2N': {'H': 4, 'N': 2}, 'OCO': {'O': 2}, 'N30': {'H': 4, 'N': 1},
    'N31': {'H': 3, 'N': 1}, 'N32': {'H': 2, 'N': 1}, 'N33': {'H': 1, 'N': 1},
    'NP1': {'H': 2, 'N': 1}, 'N1': {'N': 1}}
EXPECTED_ATOMS_BASE_INTERACTIONS = {
    'COO': {'O': 2}, 'HIS': {'N': 2}, 'CYS': {'S': 1}, 'TYR': {'O': 1},
    'LYS': {'N': 1}, 'ARG': {'N': 3}, 'ROH': {'O': 1}, 'AMD': {'O': 1},
    'TRP': {'N': 1}, 'N+': {'N': 1}, 'C-': {'O': 2}, 'BBN': {'H': 1, 'N': 1},
    'BBC': {'O': 1}, 'NAR': {'H': 1, 'N': 1}, 'NAM': {'H': 1, 'N': 1},
    'F': {'F': 1}, 'Cl': {'Cl': 1}, 'OH': {'H': 1, 'O': 1}, 'OP': {'O': 1},
    'O3': {'O': 1}, 'O2': {'O': 1}, 'SH': {'S': 1}, 'CG': {'N': 3},
    'C2N': {'N': 2}, 'OCO': {'O': 2}, 'N30': {'N': 1}, 'N31': {'N': 1},
    'N32': {'N': 1}, 'N33': {'N': 1}, 'NP1': {'N': 1}, 'N1': {'N': 1}}


class Group:
    """Class for storing groups important to pKa calculations."""

    def __init__(self, atom):
        """Initialize with an atom.

        Args:
            atom:  atom object
        """
        self.atom = atom
        self.type = ''
        atom.group = self
        # set up data structures
        self.determinants = {'sidechain': [], 'backbone': [], 'coulomb': []}
        self.pka_value = 0.0
        self.model_pka = 0.0
        # Energy associated with volume interactions
        self.energy_volume = 0.0
        # Number of atoms associated with volume interactions
        self.num_volume = 0.0
        # Energy associated with local interactions
        self.energy_local = 0.0
        # Number of atoms associated with local interactions
        self.num_local = 0.0
        self.buried = 0.0
        self.x = 0.0
        self.y = 0.0
        self.z = 0.0
        self.charge = 0
        self.parameters = None
        self.exclude_cys_from_results = None
        self.interaction_atoms_for_acids = []
        self.interaction_atoms_for_bases = []
        self.model_pka_set = False
        self.intrinsic_pka = None
        self.titratable = None
        # information on covalent and non-covalent coupling
        self.non_covalently_coupled_groups = []
        self.covalently_coupled_groups = []
        self.coupled_titrating_group = None
        self.common_charge_centre = False
        self.residue_type = self.atom.res_name
        if self.atom.terminal:
            self.residue_type = self.atom.terminal
        if self.atom.type == 'atom':
            fmt = "{g.residue_type:<3s}{a.res_num:>4d}{a.chain_id:>2s}"
            self.label = fmt.format(g=self, a=atom)
        elif self.atom.res_name in ['DA ', 'DC ', 'DG ', 'DT ']:
            fmt = "{type:1s}{elem:1s}{name:1s}{res_num:>4d}{chain:>2s}"
            self.label = fmt.format(
                type=self.residue_type[1], elem=atom.element,
                name=atom.name.replace('\'', '')[-1], res_num=atom.res_num,
                chain=atom.chain_id)
        else:
            fmt = "{type:<3s}{name:>4s}{chain:>2s}"
            self.label = fmt.format(
                type=self.residue_type, name=atom.name, chain=atom.chain_id)
        # container for squared distances
        self.squared_distances = {}

    def couple_covalently(self, other):
        """Couple this group with another group.

        Args:
            other:  other group for coupling
        """
        # do the coupling
        if not other in self.covalently_coupled_groups:
            self.covalently_coupled_groups.append(other)
        if not self in other.covalently_coupled_groups:
            other.covalently_coupled_groups.append(self)

    def couple_non_covalently(self, other):
        """Non-covalenthly couple this group with another group.

        Args:
            other:  other group for coupling
        """
        # do the coupling
        if not other in self.non_covalently_coupled_groups:
            self.non_covalently_coupled_groups.append(other)
        if not self in other.non_covalently_coupled_groups:
            other.non_covalently_coupled_groups.append(self)

    def get_covalently_coupled_groups(self):
        """Get covalently coupled groups.

        Returns:
            list of covalently coupled groups.
        """
        return self.covalently_coupled_groups

    def get_non_covalently_coupled_groups(self):
        """Get non-covalently coupled groups.

        Returns:
            list of covalently coupled groups.
        """
        return self.non_covalently_coupled_groups

    def share_determinants(self, others):
        """Share determinants between this group and others.

        Args:
            others:  list of other groups
        """
        # for each determinant type
        for other in others:
            if other == self:
                the_other = other
                continue
            for type_ in ['sidechain', 'backbone', 'coulomb']:
                for det in other.determinants[type_]:
                    self.share_determinant(det, type_)
        # recalculate pka values
        self.calculate_total_pka()
        the_other.calculate_total_pka()

    def share_determinant(self, new_determinant, type_):
        """Add determinant to this group's list of determinants.

        Args:
            new_determinant:  determinant to add
            type_:  type of determinant
        """
        added = False
        # first check if we already have a determinant with this label
        for own_determinant in self.determinants[type_]:
            if own_determinant.group == new_determinant.group:
                # if so, find the average value
                avr = 0.5*(own_determinant.value + new_determinant.value)
                own_determinant.value = avr
                new_determinant.value = avr
                added = True
        # otherwise we just add the determinant to our list
        if not added:
            self.determinants[type_].append(
                Determinant(new_determinant.group, new_determinant.value))

    def make_covalently_coupled_line(self):
        """Create line for covalent coupling.

        Returns:
            string
        """
        # first check if there are any coupled groups at all
        if len(self.covalently_coupled_groups) == 0:
            return ''
        line = 'CCOUPL{0:5d}'.format(self.atom.numb)
        # extract and sort numbers of coupled groups
        coupled = []
        for group in self.covalently_coupled_groups:
            coupled.append(group.atom.numb)
        coupled.sort()
        # write 'em out
        for num in coupled:
            line += '{0:5d}'.format(num)
        line += '\n'
        return line

    def make_non_covalently_coupled_line(self):
        """Create line for non-covalent coupling.

        Returns:
            string
        """
        # first check if there are any coupled groups at all
        if len(self.non_covalently_coupled_groups) == 0:
            return ''
        line = 'NCOUPL{0:5d}'.format(self.atom.numb)
        # extract and sort numbers of coupled groups
        coupled = []
        for group in self.non_covalently_coupled_groups:
            coupled.append(group.atom.numb)
        coupled.sort()
        # write 'em out
        for num in coupled:
            line += '{0:5d}'.format(num)
        line += '\n'
        return line

    def __eq__(self, other):
        """Needed for creating sets of groups."""
        if self.atom.type == 'atom':
            # In case of protein atoms we trust the labels
            return self.label == other.label
        else:
            # For heterogene atoms we also need to check the residue number
            return (
                (self.label == other.label)
                and (self.atom.res_num == other.atom.res_num))

    def __hash__(self):
        """Needed for creating sets of groups."""
        return id(self)

    def __iadd__(self, other):
        if self.type != other.type:
            str_ = (
                'Cannot add groups of different types '
                '({0:s} and {1:s})'.format(self.type, other.type))
            raise Exception(str_)
        # add all values
        self.pka_value += other.pka_value
        self.num_volume += other.num_volume
        self.energy_volume += other.energy_volume
        self.num_local += other.num_local
        self.energy_local += other.energy_local
        self.buried += other.buried
        # and add all determinants
        # TODO - list ['sidechain', 'backbone', 'coulomb'] should be constant
        # This list appears all over the code and should be moved to a constant
        # higher in the package
        for type_ in ['sidechain', 'backbone', 'coulomb']:
            for determinant in other.determinants[type_]:
                self.add_determinant(determinant, type_)
        return self

    def add_determinant(self, new_determinant, type_):
        """Add to current and creates non-present determinants.

        Args:
            new_determinant:  new determinant to add
            type_:  determinant type
        """
        # first check if we already have a determinant with this label
        for own_determinant in self.determinants[type_]:
            if own_determinant.group == new_determinant.group:
                # if so, add the value
                own_determinant.value += new_determinant.value
                return
        # otherwise we just add the determinant to our list
        self.determinants[type_].append(Determinant(new_determinant.group,
                                                    new_determinant.value))

    def set_determinant(self, new_determinant, type_):
        """Overwrite current and create non-present determinants.

        Args:
            new_determinant:  new determinant to add
            type_:  determinant type
        """
        # first check if we already have a determinant with this label
        for own_determinant in self.determinants[type_]:
            if own_determinant.group == new_determinant.group:
                # if so, overwrite the value
                own_determinant.value = new_determinant.value
                return
        # otherwise we just add the determinant to our list
        self.determinants[type_].append(Determinant(new_determinant.group,
                                                    new_determinant.value))

    def remove_determinants(self, labels):
        """Remove all determinants with specified labels.

        Args:
            labels:  list of labels to remove
        """
        for type_ in ['sidechain', 'backbone', 'coulomb']:
            matches = list(
                filter(lambda d: d.label
                       in labels, [d for d in self.determinants[type_]]))
            for match in matches:
                self.determinants[type_].remove(match)

    def __truediv__(self, value):
        value = float(value)
        # divide all values
        self.pka_value /= value
        self.num_volume /= value
        self.energy_volume /= value
        self.num_local /= value
        self.energy_local /= value
        self.buried /= value
        # and all determinants
        for type_ in ['sidechain', 'backbone', 'coulomb']:
            for determinant in self.determinants[type_]:
                determinant.value /= value
        return self

    def clone(self):
        """Create a copy of this group.

        Returns:
            Copy of this group.
        """
        res = Group(self.atom)
        res.type = self.type
        res.residue_type = self.residue_type
        res.model_pka = self.model_pka
        res.coupled_titrating_group = self.coupled_titrating_group
        res.covalently_coupled_groups = self.covalently_coupled_groups
        res.non_covalently_coupled_groups = self.non_covalently_coupled_groups
        res.titratable = self.titratable
        res.exclude_cys_from_results = self.exclude_cys_from_results
        res.charge = self.charge
        return res

    def setup(self):
        """Set up a group."""
        # set the charges
        if self.type in self.parameters.charge.keys():
            self.charge = self.parameters.charge[self.type]
        if self.residue_type in self.parameters.ions.keys():
            self.charge = self.parameters.ions[self.residue_type]
        # find the center and the interaction atoms
        self.setup_atoms()
        # set the model pka value
        self.titratable = False
        if self.residue_type in self.parameters.model_pkas.keys():
            if not self.model_pka_set:
                self.model_pka = self.parameters.model_pkas[self.residue_type]
                # check if we should apply a custom model pka
                key = '{0:s}-{1:s}'.format(
                    self.atom.res_name.strip(),
                    self.atom.name.strip())
                if key in self.parameters.custom_model_pkas.keys():
                    self.model_pka = self.parameters.custom_model_pkas[key]
                self.model_pka_set = True
        if self.model_pka_set and not self.atom.cysteine_bridge:
            self.titratable = True
        self.exclude_cys_from_results = False

    def setup_atoms(self):
        """Set up atoms in group.

        This method is overwritten for some types of groups
        """
        # set the center at the position of the main atom
        self.set_center([self.atom])
        # set the main atom as interaction atom
        self.set_interaction_atoms([self.atom], [self.atom])

    def set_interaction_atoms(self, interaction_atoms_for_acids,
                              interaction_atoms_for_bases):
        """Set interacting atoms and group types.

        Args:
            interaction_atoms_for_acids:  list of atoms for acid interactions
            interaction_atoms_for_base:  list of atoms for base interactions
        """
        for atom in interaction_atoms_for_acids + interaction_atoms_for_bases:
            atom.set_group_type(self.type)
        self.interaction_atoms_for_acids = interaction_atoms_for_acids
        self.interaction_atoms_for_bases = interaction_atoms_for_bases
        # check if all atoms have been identified
        ok = True
        for [expect, found, _] in [[EXPECTED_ATOMS_ACID_INTERACTIONS,
                                    self.interaction_atoms_for_acids, 'acid'],
                                   [EXPECTED_ATOMS_BASE_INTERACTIONS,
                                    self.interaction_atoms_for_bases, 'base']]:
            if self.type in expect.keys():
                for elem in expect[self.type].keys():
                    if (len([a for a in found if a.element == elem])
                            != expect[self.type][elem]):
                        ok = False
        if not ok:
            str_ = 'Missing atoms or failed protonation for '
            str_ += '{0:s} ({1:s}) -- please check the structure'.format(
                self.label, self.type)
            warning(str_)
            warning('{0:s}'.format(str(self)))
            num_acid = sum(
                [EXPECTED_ATOMS_ACID_INTERACTIONS[self.type][e]
                 for e in EXPECTED_ATOMS_ACID_INTERACTIONS[self.type].keys()])
            num_base = sum(
                [EXPECTED_ATOMS_BASE_INTERACTIONS[self.type][e]
                 for e in EXPECTED_ATOMS_BASE_INTERACTIONS[self.type].keys()])
            warning(
                'Expected {0:d} interaction atoms for acids, found:'.format(
                    num_acid))
            for i in range(len(self.interaction_atoms_for_acids)):
                warning(
                    '             {0:s}'.format(
                        str(self.interaction_atoms_for_acids[i])))
            warning(
                'Expected {0:d} interaction atoms for bases, found:'.format(
                    num_base))
            for i in range(len(self.interaction_atoms_for_bases)):
                warning(
                    '             {0:s}'.format(
                        str(self.interaction_atoms_for_bases[i])))

    def get_interaction_atoms(self, interacting_group):
        """Get atoms involved in interaction with other group.

        Args:
            interacting_group:  other group
        Returns:
            list of atoms
        """
        if interacting_group.residue_type in self.parameters.base_list:
            return self.interaction_atoms_for_bases
        else:
            # default is acid interaction atoms - cf. 3.0
            return self.interaction_atoms_for_acids

    def set_center(self, atoms):
        """Set center of group based on atoms.

        Args:
            atoms:  list of atoms
        """
        if not atoms:
            raise ValueError("At least one atom must be specified")
        # reset center
        self.x = 0.0
        self.y = 0.0
        self.z = 0.0
        # find the average position of atoms
        for atom in atoms:
            self.x += atom.x
            self.y += atom.y
            self.z += atom.z
        self.x /= float(len(atoms))
        self.y /= float(len(atoms))
        self.z /= float(len(atoms))

    def get_determinant_string(self, remove_penalised_group=False):
        """Create a string to identify this determinant.

        Args:
            remove_penalised_group:  Boolean flag to remove penalized groups
        Returns:
            string
        """
        if self.coupled_titrating_group and remove_penalised_group:
            return ''
        number_of_sidechain = len(self.determinants['sidechain'])
        number_of_backbone = len(self.determinants['backbone'])
        number_of_coulomb = len(self.determinants['coulomb'])
        number_of_lines = max(1, number_of_sidechain, number_of_backbone,
                              number_of_coulomb)
        str_ = ""
        for line_number in range(number_of_lines):
            str_ += "{0:s}".format(self.label)
            if line_number == 0:
                str_ += " {0:6.2f}".format(self.pka_value)
                if len(self.non_covalently_coupled_groups) > 0:
                    str_ += '*'
                else:
                    str_ += ' '
                str_ += " {0:4d}{1:>2s} ".format(int(100.0*self.buried), "%")
                str_ += " {0:6.2f} {1:4d}".format(
                    self.energy_volume, int(self.num_volume))
                str_ += " {0:6.2f} {1:4d}".format(
                    self.energy_local, int(self.num_local))
            else:
                str_ += "{0:>40s}".format(" ")
            # add the determinants
            for type_ in ['sidechain', 'backbone', 'coulomb']:
                str_ += self.get_determinant_for_string(type_, line_number)
            # adding end-of-line
            str_ += "\n"
        str_ += "\n"
        return str_

    def get_determinant_for_string(self, type_, number):
        """Return a string describing determinant.

        Args:
            type_:  determinant type
            number:  determinant index number
        Returns:
            string
        """
        if number >= len(self.determinants[type_]):
            return "    0.00 XXX   0 X"
        else:
            determinant = self.determinants[type_][number]
            return "{0:8.2f} {1:s}".format(
                determinant.value, determinant.label)

    def calculate_total_pka(self):
        """Calculate total pKa based on determinants associated with this
        group."""
        # if this is a cysteine involved in a di-sulphide bond
        if self.atom.cysteine_bridge:
            self.pka_value = 99.99
            return
        self.pka_value = (
            self.model_pka + self.energy_volume + self.energy_local)
        for determinant_type in ['sidechain', 'backbone', 'coulomb']:
            for determinant in self.determinants[determinant_type]:
                self.pka_value += determinant.value

    def calculate_intrinsic_pka(self):
        """Calculate the intrinsic pKa values from the desolvation
        determinants, back-bone hydrogen bonds, and side-chain hydrogen bonds
        to non-titratable residues.
        """
        back_bone = 0.0
        for determinant in self.determinants['backbone']:
            value = determinant.value
            back_bone += value
        side_chain = 0.0
        for determinant in self.determinants['sidechain']:
            if determinant.label[0:3] not in [
                    'ASP', 'GLU', 'LYS', 'ARG', 'HIS', 'CYS', 'TYR', 'C- ',
                    'N+ ']:
                value = determinant.value
                side_chain += value
        self.intrinsic_pka = (
            self.model_pka + self.energy_volume + self.energy_local
            + back_bone + side_chain)

    def get_summary_string(self, remove_penalised_group=False):
        """Create summary string for this group.

        Args:
            remove_penalised_group:  Boolean to ignore penalized groups
        Returns:
            string
        """
        if self.coupled_titrating_group and remove_penalised_group:
            return ''
        ligand_type = ''
        if self.atom.type == 'hetatm':
            ligand_type = self.type
        penalty = ''
        if self.coupled_titrating_group:
            penalty = (
                ' NB: Discarded due to coupling with {0:s}'.format(
                    self.coupled_titrating_group.label))
        fmt = (
            "   {g.label:>9s} {g.pka_value:8.2f} {g.model_pka:10.2f} "
            "{type:>18s}   {penalty:s}\n")
        return fmt.format(g=self, type=ligand_type, penalty=penalty)

    def __str__(self):
        str_ = 'Group ({0:s}) for {1:s}'.format(self.type, str(self.atom))
        return str_

    def calculate_folding_energy(self, parameters, ph=None, reference=None):
        """Return the electrostatic energy of this residue at specified pH.

        Args:
            parameters:  parameters for energy calculation
            ph:  pH value for calculation
            reference:  reference state for calculation
        Returns:
            float describing energy
        """
        if ph is None:
            ph = parameters.pH
        if reference is None:
            reference = parameters.reference
        # If not titratable, the contribution is zero
        if not self.titratable:
            return 0.00
        # calculating the ddg(neutral --> low-pH) contribution
        ddg_neutral = 0.00
        if reference == 'neutral' and self.charge > 0.00:
            pka_prime = self.pka_value
            for determinant in self.determinants['coulomb']:
                if determinant.value > 0.00:
                    pka_prime -= determinant.value
            ddg_neutral = UNK_PKA_SCALING*(pka_prime - self.model_pka)
        # calculating the ddg(low-pH --> pH) contribution
        # folded
        dpka = ph - self.pka_value
        conc_ratio = 10**dpka
        q_pro = math.log10(1+conc_ratio)
        # unfolded
        dpka = ph - self.model_pka
        conc_ratio = 10**dpka
        q_mod = math.log10(1+conc_ratio)
        ddg_low = UNK_PKA_SCALING*(q_pro - q_mod)
        ddg = ddg_neutral + ddg_low
        return ddg

    def calculate_charge(self, _, ph=7.0, state='folded'):
        """Calculate the charge of the specified state at the specified pH.

        Args:
            _:  parameters for calculation
            ph:  pH value
            state:  "folded" or "unfolded"
        Returns:
            float with charge
        """
        if state == "unfolded":
            q_dpka = self.charge * (self.model_pka - ph)
        else:
            q_dpka = self.charge * (self.pka_value - ph)
        conc_ratio = 10**q_dpka
        charge = self.charge*(conc_ratio/(1.0+conc_ratio))
        return charge

    def use_in_calculations(self):
        """Indicate whether group should be included in results report.

        If --titrate_only option is specified, only residues that are
        titratable and are in that list are included; otherwise all titratable
        residues and CYS residues are included.
        """
        return self.titratable or (
            self.residue_type == 'CYS' and not self.exclude_cys_from_results)


class COOGroup(Group):
    """Carboxyl group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'COO'

    def setup_atoms(self):
        """Set up group."""
        # Identify the two caroxyl oxygen atoms
        the_oxygens = self.atom.get_bonded_elements('O')
        # set the center using the two oxygen carboxyl atoms (if present)
        if the_oxygens:
            self.set_center(the_oxygens)
        else:
            self.set_center([self.atom])
            # TODO - perhaps it would be better to ignore this group completely
            # if the oxygen is missing from this residue?
        self.set_interaction_atoms(the_oxygens, the_oxygens)


class HISGroup(Group):
    """Histidine group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'HIS'

    def setup_atoms(self):
        """Set up atoms in group."""
        # Find the atoms in the histidine ring
        ring_atoms = propka.ligand.is_ring_member(self.atom)
        if len(ring_atoms) != 5:
            warning('His group does not seem to contain a ring', self)
        # protonate ring
        for ring_atom in ring_atoms:
            PROTONATOR.protonate_atom(ring_atom)
        # set the center using the ring atoms
        if ring_atoms:
            self.set_center(ring_atoms)
        else:
            # Missing side-chain atoms
            self.set_center([self.atom])
            # TODO - perhaps it would be better to ignore this group completely?
        # find the hydrogens on the ring-nitrogens
        hydrogens = []
        nitrogens = [ra for ra in ring_atoms if ra.element == 'N']
        for nitrogen in nitrogens:
            hydrogens.extend(nitrogen.get_bonded_elements('H'))
        self.set_interaction_atoms(hydrogens+nitrogens, nitrogens)


class CYSGroup(Group):
    """Cysteine group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'CYS'


class TYRGroup(Group):
    """Tyrosine group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'TYR'


class LYSGroup(Group):
    """Lysine group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'LYS'


class ARGGroup(Group):
    """Arginine group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'ARG'

    def setup_atoms(self):
        """Set up group."""
        # set the center at the position of the main atom
        self.set_center([self.atom])
        # find all the hydrogens on the nitrogen atoms
        nitrogens = self.atom.get_bonded_elements('N')
        for nitrogen in nitrogens:
            PROTONATOR.protonate_atom(nitrogen)
        hydrogens = []
        for nitrogen in nitrogens:
            hydrogens.extend(nitrogen.get_bonded_elements('H'))
        self.set_interaction_atoms(nitrogens+hydrogens, nitrogens)


class ROHGroup(Group):
    """Alcohol group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'ROH'


class SERGroup(Group):
    """Serine group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'SER'


class AMDGroup(Group):
    """Amide group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'AMD'

    def setup_atoms(self):
        """Setup group."""
        # Identify the oxygen and nitrogen amide atoms
        the_oxygen = self.atom.get_bonded_elements('O')
        the_nitrogen = self.atom.get_bonded_elements('N')
        # add protons to the nitrogen
        PROTONATOR.protonate_atom(the_nitrogen[0])
        the_hydrogens = the_nitrogen[0].get_bonded_elements('H')
        # set the center using the oxygen and nitrogen amide atoms
        self.set_center(the_oxygen+the_nitrogen)
        self.set_interaction_atoms(the_nitrogen + the_hydrogens, the_oxygen)


class TRPGroup(Group):
    """Tryptophan group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'TRP'

    def setup_atoms(self):
        """Set up atoms in group."""
        # set the center at the position of the main atom
        self.set_center([self.atom])
        # find the hydrogen on the nitrogen atom
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogen = self.atom.get_bonded_elements('H')
        self.set_interaction_atoms(the_hydrogen+[self.atom], [self.atom])


class NtermGroup(Group):
    """N-terminus group."""
    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'N+'


class CtermGroup(Group):
    """C-terminus group."""
    def __init__(self, atom):
        Group.__init__(self, atom)
        # this is to deal with the COO-C- parameter unification.
        self.type = 'COO'

    def setup_atoms(self):
        """Set up atoms in group."""
        # Identify the carbon and other oxygen carboxyl atoms
        the_carbons = self.atom.get_bonded_elements('C')
        if not the_carbons:
            self.set_center([self.atom])
            # TODO - perhaps it would be better to ignore this group completely
            # if the carbon is missing from this residue?
        else:
            the_other_oxygen = the_carbons[0].get_bonded_elements('O')
            the_other_oxygen.remove(self.atom)
            # set the center and interaction atoms
            the_oxygens = [self.atom]+ the_other_oxygen
            self.set_center(the_oxygens)
            self.set_interaction_atoms(the_oxygens, the_oxygens)


class BBNGroup(Group):
    """Backbone nitrogen group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'BBN'
        self.residue_type = 'BBN'

    def setup_atoms(self):
        """Set up atoms in group."""
        # Identify the hydrogen
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogen = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(
            the_hydrogen+[self.atom], the_hydrogen+[self.atom])


class BBCGroup(Group):
    """Backbone carbon group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'BBC'
        self.residue_type = 'BBC'

    def setup_atoms(self):
        """Set up atoms in group."""
        # Identify the oxygen
        the_oxygen = self.atom.get_bonded_elements('O')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(the_oxygen, the_oxygen)


class NARGroup(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'NAR'
        self.residue_type = 'NAR'
        info('Found NAR group:', atom)

    def setup_atoms(self):
        """Set up atoms in group."""
        # Identify the hydrogen
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogen = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(
            the_hydrogen+[self.atom], the_hydrogen+[self.atom])


class NAMGroup(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'NAM'
        self.residue_type = 'NAM'
        info('Found NAM group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the hydrogen
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogen = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(
            the_hydrogen+[self.atom], the_hydrogen+[self.atom])


class FGroup(Group):
    """Fluoride group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'F'
        self.residue_type = 'F'
        info('Found F   group:', atom)


class ClGroup(Group):
    """Chloride group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'Cl'
        self.residue_type = 'Cl'
        info('Found Cl   group:', atom)


class OHGroup(Group):
    """Hydroxide group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'OH'
        self.residue_type = 'OH'
        info('Found OH  group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the hydrogen
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogen = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(
            the_hydrogen+[self.atom], the_hydrogen+[self.atom])


class OPGroup(Group):
    """Phosphate group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'OP'
        self.residue_type = 'OP'
        info('Found OP  group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the hydrogen
        PROTONATOR.protonate_atom(self.atom)
        # set the center using the oxygen
        self.set_center([self.atom])
        self.set_interaction_atoms([self.atom], [self.atom])


class O3Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'O3'
        self.residue_type = 'O3'
        info('Found O3  group:', atom)


class O2Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'O2'
        self.residue_type = 'O2'
        info('Found O2  group:', atom)


class SHGroup(Group):
    """Sulfhydryl group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'SH'
        self.residue_type = 'SH'
        info('Found SH  group:', atom)


class CGGroup(Group):
    """Guadinium group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'CG'
        self.residue_type = 'CG'
        info('Found CG  group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the nitrogens
        the_nitrogens = self.atom.get_bonded_elements('N')
        # set the center using the nitrogen
        self.set_center([self.atom])
        the_hydrogens = []
        for nitrogen in the_nitrogens:
            PROTONATOR.protonate_atom(nitrogen)
            the_hydrogens += nitrogen.get_bonded_elements('H')
        self.set_interaction_atoms(
            the_hydrogens+the_nitrogens, the_nitrogens)


class C2NGroup(Group):
    """Amidinium group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'C2N'
        self.residue_type = 'C2N'
        info('Found C2N group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the nitrogens
        the_nitrogens = self.atom.get_bonded_elements('N')
        the_nitrogens = [
            n for n in the_nitrogens if len(n.get_bonded_heavy_atoms()) == 1]
        # set the center using the nitrogen
        self.set_center([self.atom])
        the_hydrogens = []
        for nitrogen in the_nitrogens:
            PROTONATOR.protonate_atom(nitrogen)
            the_hydrogens += nitrogen.get_bonded_elements('H')
        self.set_interaction_atoms(the_hydrogens+the_nitrogens, the_nitrogens)


class OCOGroup(Group):
    """Carboxyl group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'OCO'
        self.residue_type = 'OCO'
        info('Found OCO group:', atom)

    def setup_atoms(self):
        """Set up atoms in group."""
        # Identify the two carboxyl oxygen atoms
        the_oxygens = self.atom.get_bonded_elements('O')
        # set the center using the two oxygen carboxyl atoms
        self.set_center(the_oxygens)
        self.set_interaction_atoms(the_oxygens, the_oxygens)


class N30Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'N30'
        self.residue_type = 'N30'
        info('Found N30 group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the nitrogens
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogens = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(the_hydrogens+[self.atom], [self.atom])


class N31Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'N31'
        self.residue_type = 'N31'
        info('Found N31 group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the nitrogens
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogens = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(the_hydrogens+[self.atom], [self.atom])


class N32Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'N32'
        self.residue_type = 'N32'
        info('Found N32 group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the nitrogens
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogens = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(the_hydrogens+[self.atom], [self.atom])


class N33Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'N33'
        self.residue_type = 'N33'
        info('Found N33 group:', atom)

    def setup_atoms(self):
        """Set up atoms in this group."""
        # Identify the nitrogens
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogens = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(the_hydrogens+[self.atom], [self.atom])


class NP1Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'NP1'
        self.residue_type = 'NP1'
        info('Found NP1 group:', atom)

    def setup_atoms(self):
        """Set up atoms in group."""
        # Identify the nitrogens
        PROTONATOR.protonate_atom(self.atom)
        the_hydrogens = self.atom.get_bonded_elements('H')
        # set the center using the nitrogen
        self.set_center([self.atom])
        self.set_interaction_atoms(the_hydrogens+[self.atom], [self.atom])


class N1Group(Group):
    """Unknown group.

    TODO - identify this group.
    """

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'N1'
        self.residue_type = 'N1'
        info('Found N1 group:', atom)


class IonGroup(Group):
    """Ion group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'ION'
        self.residue_type = atom.res_name.strip()
        info('Found ion group:', atom)


class NonTitratableLigandGroup(Group):
    """Non-titratable ligand group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        self.type = 'LG'
        self.residue_type = 'LG'


class TitratableLigandGroup(Group):
    """Titratable ligand group."""

    def __init__(self, atom):
        Group.__init__(self, atom)
        # set the charge and determine type (acid or base)
        self.charge = atom.charge
        if self.charge < 0:
            self.type = 'ALG'
            self.residue_type = 'ALG'
        elif self.charge > 0:
            self.type = 'BLG'
            self.residue_type = 'BLG'
        else:
            raise Exception('Unable to determine type of ligand group - '
                            'charge not set?')
        # check if marvin model pka has been calculated
        # this is not true if we are reading an input file
        if atom.marvin_pka:
            self.model_pka = atom.marvin_pka
            info('Titratable ligand group    ',
                 atom, self.model_pka, self.charge)
        self.model_pka_set = True


def is_group(parameters, atom):
    """Identify whether the atom belongs to a group.

    Args:
        parameters:  parameters for check
        atom:  atom to check
    Returns:
        group for atom or None
    """
    atom.groups_extracted = 1
    # check if this atom belongs to a protein group
    protein_group = is_protein_group(parameters, atom)
    if protein_group:
        return protein_group
    # check if this atom belongs to a ion group
    ion_group = is_ion_group(parameters, atom)
    if ion_group:
        return ion_group
    # check if this atom belongs to a ligand group
    if parameters.ligand_typing == 'marvin':
        ligand_group = is_ligand_group_by_marvin_pkas(parameters, atom)
    elif parameters.ligand_typing == 'sybyl':
        ligand_group = None
    elif parameters.ligand_typing == 'groups':
        ligand_group = is_ligand_group_by_groups(parameters, atom)
    else:
        raise Exception(
            'Unknown ligand typing method \'{0.s}\''.format(
                parameters.ligand_typing))
    if ligand_group:
        return ligand_group
    return None


def is_protein_group(parameters, atom):
    """Identify whether the atom belongs to a protein group.

    Args:
        parameters:  parameters for check
        atom:  atom to check
    Returns:
        group for atom or None
    """
    if atom.type != 'atom':
        return None
    ### Check for termial groups
    if atom.terminal == 'N+':
        return NtermGroup(atom)
    elif atom.terminal == 'C-':
        return CtermGroup(atom)
    ### Backbone
    if atom.type == 'atom' and atom.name == 'N':
        # ignore proline backbone nitrogens
        if atom.res_name != 'PRO':
            return BBNGroup(atom)
    if atom.type == 'atom' and atom.name == 'C':
        # ignore C- carboxyl
        if atom.count_bonded_elements('O') == 1:
            return BBCGroup(atom)
    ### Filters for side chains based on PDB protein atom names
    key = '{0:s}-{1:s}'.format(atom.res_name, atom.name)
    if key in parameters.protein_group_mapping.keys():
        class_str = "{0:s}Group".format(parameters.protein_group_mapping[key])
        group_class = globals()[class_str]
        return group_class(atom)
    return None


def is_ligand_group_by_groups(_, atom):
    """Identify whether the atom belongs to a ligand group by checking groups.

    Args:
        _:  parameters for check
        atom:  atom to check
    Returns:
        group for atom or None
    """
    ### Ligand group filters
    if atom.type != 'hetatm':
        return None
    PROTONATOR.protonate_atom(atom)
    if atom.sybyl_type == 'N.ar':
        if len(atom.get_bonded_heavy_atoms()) == 2:
            return NARGroup(atom)
    if atom.sybyl_type == 'N.am':
        return NAMGroup(atom)
    if atom.sybyl_type in ['N.3', 'N.4']:
        heavy_bonded = atom.get_bonded_heavy_atoms()
        if len(heavy_bonded) == 0:
            return N30Group(atom)
        elif len(heavy_bonded) == 1:
            return N31Group(atom)
        elif len(heavy_bonded) == 2:
            return N32Group(atom)
        elif len(heavy_bonded) == 3:
            return N33Group(atom)
    if atom.sybyl_type == 'N.1':
        return N1Group(atom)
    if atom.sybyl_type == 'N.pl3':
        # make sure that this atom is not part of a guadinium or amidinium group
        bonded_carbons = atom.get_bonded_elements('C')
        if len(bonded_carbons) == 1:
            bonded_nitrogens = bonded_carbons[0].get_bonded_elements('N')
            if len(bonded_nitrogens) == 1:
                return NP1Group(atom)
    if atom.sybyl_type == 'C.2':
        # Guadinium and amidinium groups
        bonded_nitrogens = atom.get_bonded_elements('N')
        npls = [
            n for n in bonded_nitrogens
            if (n.sybyl_type == 'N.pl3'
                and len(n.get_bonded_heavy_atoms()) == 1)]
        if len(npls) == 2:
            n_with_max_two_heavy_atom_bonds = [
                n for n in bonded_nitrogens
                if len(n.get_bonded_heavy_atoms()) < 3]
            if len(n_with_max_two_heavy_atom_bonds) == 2:
                return C2NGroup(atom)
            if len(bonded_nitrogens) == 3:
                return CGGroup(atom)
        # carboxyl group
        bonded_oxygens = atom.get_bonded_elements('O')
        bonded_oxygens = [b for b in bonded_oxygens if 'O.co2' in b.sybyl_type]
        if len(bonded_oxygens) == 2:
            return OCOGroup(atom)
    if atom.sybyl_type == 'F':
        return FGroup(atom)
    if atom.sybyl_type == 'Cl':
        return ClGroup(atom)
    if atom.sybyl_type == 'O.3':
        if len(atom.get_bonded_heavy_atoms()) == 1:
            # phosphate group
            if atom.count_bonded_elements('P') == 1:
                return OPGroup(atom)
            # hydroxyl group
            else:
                return OHGroup(atom)
        # other SP3 oxygen
        else:
            return O3Group(atom)
    if atom.sybyl_type == 'O.2':
        return O2Group(atom)
    if atom.sybyl_type == 'S.3':
        # sulphide group
        if len(atom.get_bonded_heavy_atoms()) == 1:
            return SHGroup(atom)
    return None


def is_ligand_group_by_marvin_pkas(parameters, atom):
    """Identify whether the atom belongs to a ligand group by calculating
    'Marvin pKas'.

    Args:
        parameters:  parameters for check
        atom:  atom to check
    Returns:
        group for atom or None
    """
    if atom.type != 'hetatm':
        return None
    # calculate Marvin ligand pkas for this conformation container
    # if not already done
    # TODO - double-check testing coverage of these functions.
    if not atom.conformation_container.marvin_pkas_calculated:
        lpka = LigandPkaValues(parameters)
        lpka.get_marvin_pkas_for_molecular_container(
            atom.molecular_container, min_ph=parameters.min_ligand_model_pka,
            max_ph=parameters.max_ligand_model_pka)
    if atom.marvin_pka:
        return TitratableLigandGroup(atom)
    # Special case of oxygen in carboxyl group not assigned pka value by marvin
    if atom.sybyl_type == 'O.co2':
        atom.charge = -1.0
        other_oxygen = [
            o for o
            in atom.get_bonded_elements('C')[0].get_bonded_elements('O')
            if not o == atom][0]
        atom.marvin_pka = other_oxygen.marvin_pka
        return TitratableLigandGroup(atom)
    if atom.element in parameters.hydrogen_bonds.elements:
        return NonTitratableLigandGroup(atom)
    return None


def is_ion_group(parameters, atom):
    """Identify whether the atom belongs to an ion group.

    Args:
        parameters:  parameters for check
        atom:  atom to check
    Returns:
        group for atom or None
    """
    if atom.res_name.strip() in parameters.ions.keys():
        return IonGroup(atom)
    return None

def initialize_atom_group(atom):
    """Initialize an atom group.

    Args:
        atom:  atom to initialize
    """
        # try to initialise the group
    group_attr = globals()[atom.group_label]
    atom.group = group_attr(atom)
    atom.group.model_pka = atom.group_model_pka
    atom.group.model_pka_set = atom.group_model_pka_set



"""Vector algebra for PROPKA."""
import math
from propka.lib import info, get_sorted_configurations


class Vector:
    """Vector"""

    def __init__(self, xi=0.0, yi=0.0, zi=0.0, atom1=None, atom2=None):
        """Initialize vector.

        Args:
            xi:  default x-coordinate
            yi:  default y-coordinate
            zi:  default z-coordinate
            atom1:  atom center used to define default coordinate
            atom2:  two atom centers used to define vector
        """
        self.x = xi
        self.y = yi
        self.z = zi

        if atom1:
            # make vector pointing to atom1
            self.x = atom1.x
            self.y = atom1.y
            self.z = atom1.z

            if atom2:
                # make inter-atomic vector (atom1 -> atom2)
                self.x = atom2.x - self.x
                self.y = atom2.y - self.y
                self.z = atom2.z - self.z

    def __add__(self, other):
        return Vector(self.x + other.x,
                      self.y + other.y,
                      self.z + other.z)

    def __sub__(self, other):
        return Vector(self.x - other.x,
                      self.y - other.y,
                      self.z - other.z)

    def __mul__(self, other):
        """Dot product, scalar and matrix multiplication."""
        if isinstance(other, Vector):
            return self.x * other.x + self.y * other.y + self.z * other.z
        elif isinstance(other, Matrix4x4):
            return Vector(
                xi=other.a11*self.x + other.a12*self.y + other.a13*self.z
                + other.a14*1.0,
                yi=other.a21*self.x + other.a22*self.y + other.a23*self.z
                + other.a24*1.0,
                zi=other.a31*self.x + other.a32*self.y + other.a33*self.z
                + other.a34*1.0
                )
        elif type(other) in [int, float]:
            return Vector(self.x * other, self.y * other, self.z * other)
        else:
            info('{0:s} not supported'.format(type(other)))
            raise TypeError

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, other):
        """Cross product."""
        return Vector(self.y * other.z - self.z * other.y,
                      self.z * other.x - self.x * other.z,
                      self.x * other.y - self.y * other.x)

    def __neg__(self):
        res = Vector(xi=-self.x,
                     yi=-self.y,
                     zi=-self.z)
        return res

    def sq_length(self):
        """Return vector squared-length"""
        return self.x * self.x + self.y * self.y + self.z * self.z

    def length(self):
        """Return vector length."""
        return math.sqrt(self.sq_length())

    def __str__(self):
        return '{0:>10.4f} {1:>10.4f} {2:>10.4f}'.format(
            self.x, self.y, self.z)

    def __repr__(self):
        return '<vector>'

    def orthogonal(self):
        """ Returns a vector orthogonal to self """
        res = Vector(self.y, -self.x, 0)
        if abs(self.y) < abs(self.z):
            res = Vector(self.z, 0, -self.x)
        return res

    def rescale(self, new_length):
        """ Rescale vector to new length while preserving direction """
        frac = new_length/(self.length())
        res = Vector(xi=self.x*frac, yi=self.y*frac, zi=self.z*frac)
        return res


class Matrix4x4:
    """A 4-by-4 matrix class."""

    def __init__(self,
                 a11i=0.0, a12i=0.0, a13i=0.0, a14i=0.0,
                 a21i=0.0, a22i=0.0, a23i=0.0, a24i=0.0,
                 a31i=0.0, a32i=0.0, a33i=0.0, a34i=0.0,
                 a41i=0.0, a42i=0.0, a43i=0.0, a44i=0.0):
        """Initialize with matrix elements."""
        # Row 1
        self.a11 = a11i
        self.a12 = a12i
        self.a13 = a13i
        self.a14 = a14i
        # Row 2
        self.a21 = a21i
        self.a22 = a22i
        self.a23 = a23i
        self.a24 = a24i
        # Row 3
        self.a31 = a31i
        self.a32 = a32i
        self.a33 = a33i
        self.a34 = a34i
        # Row 4
        self.a41 = a41i
        self.a42 = a42i
        self.a43 = a43i
        self.a44 = a44i


def angle(avec, bvec):
    """Get the angle between two vectors.

    Args:
        avec:  vector 1
        bvec:  vector 2
    Returns:
        angle in radians
    """
    dot = avec * bvec
    return math.acos(dot / (avec.length() * bvec.length()))


def angle_degrees(avec, bvec):
    """Get the angle between two vectors in degrees.

    Args:
        avec:  vector 1
        bvec:  vector 2
    Returns:
        angle in degrees
    """
    return math.degrees(angle(avec, bvec))


def signed_angle_around_axis(avec, bvec, axis):
    """Get signed angle of two vectors around axis in radians.

    Args:
        avec:  vector 1
        bvec:  vector 2
        axis:  axis
    Returns:
        angle in radians
    """
    norma = avec**axis
    normb = bvec**axis
    ang = angle(norma, normb)
    dot_ = bvec*(avec**axis)
    if dot_ < 0:
        ang = -ang
    return ang


def rotate_vector_around_an_axis(theta, axis, vec):
    """Rotate vector around an axis.

    Args:
        theta:  rotation angle (in radians)
        axis:  axis for rotation
        vec:  vector to rotate
    Returns:
        rotated vector
    """
    gamma = 0.0
    if axis.y != 0:
        if axis.x != 0:
            gamma = -axis.x/abs(axis.x)*math.asin(
                axis.y/(math.sqrt(axis.x*axis.x + axis.y*axis.y)))
        else:
            gamma = math.pi/2.0
        rot_z = rotate_atoms_around_z_axis(gamma)
        vec = rot_z * vec
        axis = rot_z * axis
    beta = 0.0
    if axis.x != 0:
        beta = -axis.x/abs(axis.x)*math.acos(
            axis.z/math.sqrt(axis.x*axis.x + axis.z*axis.z))
        rot_y = rotate_atoms_around_y_axis(beta)
        vec = rot_y * vec
        axis = rot_y * axis
    rot_z = rotate_atoms_around_z_axis(theta)
    vec = rot_z * vec
    rot_y = rotate_atoms_around_y_axis(-beta)
    vec = rot_y * vec
    rot_z = rotate_atoms_around_z_axis(-gamma)
    vec = rot_z * vec
    return vec


def rotate_atoms_around_z_axis(theta):
    """Get rotation matrix for z-axis.

    Args:
        theta:  angle of rotation (radians)
    Returns:
        rotation matrix
    """
    return Matrix4x4(
        a11i=math.cos(theta),
        a12i=-math.sin(theta),
        a13i=0.0,
        a14i=0.0,
        a21i=math.sin(theta),
        a22i=math.cos(theta),
        a23i=0.0,
        a24i=0.0,
        a31i=0.0,
        a32i=0.0,
        a33i=1.0,
        a34i=0.0,
        a41i=0.0,
        a42i=0.0,
        a43i=0.0,
        a44i=1.0
        )


def rotate_atoms_around_y_axis(theta):
    """Get rotation matrix for y-axis.

    Args:
        theta:  angle of rotation (radians)
    Returns:
        rotation matrix
    """
    return Matrix4x4(
        a11i=math.cos(theta),
        a12i=0.0,
        a13i=math.sin(theta),
        a14i=0.0,
        a21i=0.0,
        a22i=1.0,
        a23i=0.0,
        a24i=0.0,
        a31i=-math.sin(theta),
        a32i=0.0,
        a33i=math.cos(theta),
        a34i=0.0,
        a41i=0.0,
        a42i=0.0,
        a43i=0.0,
        a44i=1.0
        )


class MultiVector:
    """Collection of vectors for multiple configurations of atoms.

    TODO - this class does not appear to be used or covered by tests
    """

    def __init__(self, atom1=None, atom2=None):
        """Initialize with atom configurations.

        Args:
            atom1:  first atom to define vector
            atom2:  second atom to define vector
        """
        self.vectors = []
        self.keys = []
        self.result = None
        # store vectors for all configurations of atoms
        if atom1 is not None:
            self.keys = get_sorted_configurations(atom1.configurations.keys())
            if atom2 is not None:
                keys2 = get_sorted_configurations(atom2.configurations.keys())
                if self.keys != keys2:
                    str_ = ('Cannot make multi vector: Atomic configurations '
                            'mismatch for\n   {0:s}\n   {1:s}\n'.format(
                                atom1, atom2))
                    raise KeyError(str_)
            for key in self.keys:
                atom1.setConfiguration(key)
                if atom2 != 0:
                    atom2.setConfiguration(key)
                vec = Vector(atom1=atom1, atom2=atom2)
                self.vectors.append(vec)

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return self.do_job(name)

    def __str__(self):
        res = ''
        for i, key in enumerate(self.keys):
            res += '{0:s} {1:s}\n'.format(key, self.vectors[i])
        return res

    def do_job(self, job):
        """Append vectors to configuration.

        Args:
            job:  name of function to apply to vectors
        Returns:
            TODO - figure out what this is
        """
        self.result = MultiVector()
        for i, vector in enumerate(self.vectors):
            func = getattr(vector, job)
            self.result.vectors.append(func())
            self.result.keys.append(self.keys[i])
        return self.get_result

    @property
    def get_result(self):
        """Return the latest result."""
        return self.result

    def generic_operation(self, operation, other):
        """Perform a generic operation between two MultiVector objects.

        Args:
            operation:  operation to perform (string)
            other:  other MultiVector object
        """
        if self.keys != other.keys:
            raise 'Incompatible keys'
        self.result = MultiVector()
        for i in range(len(self.vectors)):
            self.result.vectors.append(
                # TODO - eliminate eval() or entire class
                eval(
                    'self.vectors[{0:d}] {1:s} other.vectors[{2:d}]'.format(
                        i, operation, i)))
            self.result.keys.append(self.keys[i])

    def __add__(self, other):
        self.generic_operation('+', other)
        return self.result

    def __sub__(self, other):
        self.generic_operation('-', other)
        return self.result

    def __mul__(self, other):
        self.generic_operation('*', other)
        return self.result

    def __pow__(self, other):
        self.generic_operation('**', other)
        return self.result

    @staticmethod
    def generic_self_operation(_):
        """TODO - delete this."""
        return

    def __neg__(self):
        self.generic_operation('*', -1.0)
        return self.result

    def rescale(self, new_length):
        """Rescale multi-vector to new length.

        Args:
            new_length:  new length for multi-vector
        Result:
            MultiVector object
        """
        self.result = MultiVector()
        for i, vector in enumerate(self.vectors):
            self.result.vectors.append(vector.rescale(new_length))
            self.result.keys.append(self.keys[i])
        return self.res


def rotate_multi_vector_around_an_axis(theta, axis, vec):
    """Rotate a multi-vector around an axis.

    NOTE - both axis ans v must be MultiVectors.

    Args:
        theta:  angle (in radians)
        axis:  multi-vector axis
        vec:  multi-vector vector
    """
    if axis.keys != vec.keys:
        raise 'Incompatible keys in rotate MultiVector'
    res = MultiVector()
    for i, key in enumerate(vec.keys):
        res.vectors.append(rotate_vector_around_an_axis(
            theta, axis.vectors[i], vec.vectors[i]))
        res.keys.append(key)
    return res



"""Input routines."""
from pathlib import Path
from pkg_resources import resource_filename
from propka.lib import protein_precheck
from propka.output import write_propka
from propka.atom import Atom
from propka.conformation_container import ConformationContainer
from propka.group import initialize_atom_group


def open_file_for_reading(input_file):
    """Open file or file-like stream for reading.

    TODO - convert this to a context manager

    Args:
        input_file: path to file or file-like object. If file-like object,
        then will attempt fseek(0).
    """
    try:
        input_file.fseek(0)
        return input_file
    except AttributeError:
        pass

    try:
        file_ = open(input_file, 'rt')
    except:
        raise IOError('Cannot find file {0:s}'.format(input_file))
    return file_


def read_molecule_file(input_file, mol_container):
    """Read input file (PDB or PROPKA) for a molecular container

    Args
        input_file:  input file to read
        mol_container:  MolecularContainer object
    Returns
        updated MolecularContainer object
    Raises
        ValuError if invalid input given
    """
    input_path = Path(input_file)
    mol_container.name = input_path.stem
    input_file_extension = input_path.suffix
    if input_file_extension.lower() == '.pdb':
        # input is a pdb file. read in atoms and top up containers to make
        # sure that all atoms are present in all conformations
        conformations, conformation_names = read_pdb(
            input_path, mol_container.version.parameters, mol_container)
        if len(conformations) == 0:
            str_ = ('Error: The pdb file does not seems to contain any '
                    'molecular conformations')
            raise ValueError(str_)
        mol_container.conformations = conformations
        mol_container.conformation_names = conformation_names
        mol_container.top_up_conformations()
        # make a structure precheck
        protein_precheck(
            mol_container.conformations, mol_container.conformation_names)
        # set up atom bonding and protonation
        mol_container.version.setup_bonding_and_protonation(mol_container)
        # Extract groups
        mol_container.extract_groups()
        # sort atoms
        for name in mol_container.conformation_names:
            mol_container.conformations[name].sort_atoms()
        # find coupled groups
        mol_container.find_covalently_coupled_groups()
    elif input_file_extension.lower() == '.propka_input':
        # input is a propka_input file
        conformations, conformation_names = read_propka(
            input_file, mol_container.version.parameters, mol_container)
        mol_container.conformations = conformations
        mol_container.conformation_names = conformation_names
        # Extract groups - this merely sets up the groups found in the
        # input file
        mol_container.extract_groups()
        # do some additional set up
        mol_container.additional_setup_when_reading_input_file()
    else:
        str_ = "Unknown input file type {0!s} for file {1!s}".format(
            input_file_extension, input_path)
        raise ValueError(str_)
    return mol_container


def read_parameter_file(input_file, parameters):
    """Read a parameter file.

    Args:
        input_file:  input file to read
        parameters:  Parameters object
    Returns:
        updated Parameters object
    """
    # try to locate the parameter file
    try:
        ifile = resource_filename(__name__, input_file)
        input_ = open_file_for_reading(ifile)
    except (IOError, FileNotFoundError, ValueError, KeyError):
        input_ = open_file_for_reading(input_file)
    for line in input_:
        parameters.parse_line(line)
    return parameters


def conformation_sorter(conf):
    """TODO - figure out what this function does."""
    model = int(conf[:-1])
    altloc = conf[-1:]
    return model*100+ord(altloc)


def get_atom_lines_from_pdb(pdb_file, ignore_residues=[], keep_protons=False,
                            tags=['ATOM  ', 'HETATM'], chains=None):
    """Get atom lines from PDB file.

    Args:
        pdb_file:  PDB file to parse
        ignore_residues:  list of residues to ignore
        keep_protons:  bool to keep/ignore protons
        tags:  tags of lines that include atoms
        chains:  list of chains
    """
    lines = open_file_for_reading(pdb_file).readlines()
    nterm_residue = 'next_residue'
    old_residue = None
    terminal = None
    model = 1
    for line in lines:
        tag = line[0:6]
        # set the model number
        if tag == 'MODEL ':
            model = int(line[6:])
            nterm_residue = 'next_residue'
        if tag == 'TER   ':
            nterm_residue = 'next_residue'
        if tag in tags:
            alt_conf_tag = line[16]
            residue_name = line[12: 16]
            residue_number = line[22: 26]
            # check if we want this residue
            if line[17: 20] in ignore_residues:
                continue
            if chains and line[21] not in chains:
                continue
            # set the Nterm residue number - nessecary because we may need to
            # identify more than one N+ group for structures with alt_conf tags
            if nterm_residue == 'next_residue' and tag == 'ATOM  ':
                # make sure that we reached a new residue - nessecary if OXT is
                # not the last atom inthe previous residue
                if old_residue != residue_number:
                    nterm_residue = residue_number
                    old_residue = None
            # Identify the configuration
            # convert digits to letters
            if alt_conf_tag in '123456789':
                alt_conf_tag = chr(ord(alt_conf_tag)+16)
            if alt_conf_tag == ' ':
                alt_conf_tag = 'A'
            conformation = '{0:d}{1:s}'.format(model, alt_conf_tag)
            # set the terminal
            if  tag == 'ATOM  ':
                if (residue_name.strip() == 'N'
                        and nterm_residue == residue_number):
                    terminal = 'N+'
                if  residue_name.strip() in ['OXT', 'O\'\'']:
                    terminal = 'C-'
                    nterm_residue = 'next_residue'
                    old_residue = residue_number
            # and yield the atom
            atom = Atom(line=line)
            atom.terminal = terminal
            #ignore hydrogen
            if not (atom.element == 'H' and not keep_protons):
                yield (conformation, atom)
            terminal = None


def read_propka(input_file, parameters, molecule):
    """Read PROPKA input file for molecular container.

    Args:
        input_file:  input file
        parameters:  parameters for parsing/setup
        molecule:  molecular container
    Returns:
        list with [conformations, names of conformations]
    """
    conformations = {}
    # read in all atoms in the input file
    lines = get_atom_lines_from_input(input_file)
    for (name, atom) in lines:
        if not name in conformations.keys():
            conformations[name] = ConformationContainer(
                name=name, parameters=parameters,
                molecular_container=molecule)
        conformations[name].add_atom(atom)
    # make a sorted list of conformation names
    names = sorted(conformations.keys(), key=conformation_sorter)
    return [conformations, names]


def get_atom_lines_from_input(input_file, tags=['ATOM  ', 'HETATM']):
    """Get atom lines from a PROPKA input file.

    Args:
        input_file:  input file
        tags:  tags defining atom lines
    Yields:
        conformation container, list of atoms
    """
    lines = open_file_for_reading(input_file).readlines()
    conformation = ''
    atoms = {}
    numbers = []
    for line in lines:
        tag = line[0:6]
        # set the conformation
        if tag == 'MODEL ':
            conformation = line[6:].strip()
        # found an atom - save it
        if tag in tags:
            atom = Atom(line=line)
            atom.get_input_parameters()
            initialize_atom_group(atom)
            atom.groups_extracted = 1
            atom.is_protonated = True
            atoms[atom.numb] = atom
            numbers.append(atom.numb)
        # found bonding information - apply it
        if tag == 'CONECT' and len(line) > 14:
            conect_numbers = [line[i:i+5] for i in range(6, len(line)-1, 5)]
            center_atom = atoms[int(conect_numbers[0])]
            for num in conect_numbers[1:]:
                bond_atom = atoms[int(num)]
                # remember to check for cysteine bridges
                if center_atom.element == 'S' and bond_atom.element == 'S':
                    center_atom.cysteine_bridge = True
                    bond_atom.cysteine_bridge = True
                # set up bonding
                if not bond_atom in center_atom.bonded_atoms:
                    center_atom.bonded_atoms.append(bond_atom)
                if not center_atom in bond_atom.bonded_atoms:
                    bond_atom.bonded_atoms.append(center_atom)
        # found info on covalent coupling
        if tag == 'CCOUPL' and len(line) > 14:
            conect_numbers = [line[i:i+5] for i in range(6, len(line)-1, 5)]
            center_atom = atoms[int(conect_numbers[0])]
            for num in conect_numbers[1:]:
                cov_atom = atoms[int(num)]
                center_atom.group.couple_covalently(cov_atom.group)
        # found info on non-covalent coupling
        if tag == 'NCOUPL' and len(line) > 14:
            conect_numbers = [line[i:i+5] for i in range(6, len(line)-1, 5)]
            center_atom = atoms[int(conect_numbers[0])]
            for num in conect_numbers[1:]:
                cov_atom = atoms[int(num)]
                center_atom.group.couple_non_covalently(cov_atom.group)
        # this conformation is done - yield the atoms
        if tag == 'ENDMDL':
            for num in numbers:
                yield (conformation, atoms[num])
            # prepare for next conformation
            atoms = {}
            numbers = []

def read_pdb(pdb_file, parameters, molecule):
    """Parse a PDB file.

    Args:
        pdb_file:  file to read
        parameters:  parameters to guide parsing
        molecule:  molecular container
    Returns:
        list with elements:
            1. list of conformations
            2. list of names
    """
    conformations = {}
    # read in all atoms in the file
    lines = get_atom_lines_from_pdb(
        pdb_file, ignore_residues=parameters.ignore_residues,
        keep_protons=molecule.options.keep_protons,
        chains=molecule.options.chains)
    for (name, atom) in lines:
        if not name in conformations.keys():
            conformations[name] = ConformationContainer(
                name=name, parameters=parameters, molecular_container=molecule)
        conformations[name].add_atom(atom)
    # make a sorted list of conformation names
    names = sorted(conformations.keys(), key=conformation_sorter)
    return [conformations, names]



"""Holds parameters and settings."""
from propka.lib import info, warning


# names and types of all key words in configuration file
MATRICES = ['interaction_matrix']
PAIR_WISE_MATRICES = ['sidechain_cutoffs']
NUMBER_DICTIONARIES = [
    'VanDerWaalsVolume', 'charge', 'model_pkas', 'ions', 'valence_electrons',
    'custom_model_pkas']
LIST_DICTIONARIES = ['backbone_NH_hydrogen_bond', 'backbone_CO_hydrogen_bond']
STRING_DICTIONARIES = ['protein_group_mapping']
STRING_LISTS = [
    'ignore_residues', 'angular_dependent_sidechain_interactions',
    'acid_list', 'base_list', 'exclude_sidechain_interactions',
    'backbone_reorganisation_list', 'write_out_order']
DISTANCES = ['desolv_cutoff', 'buried_cutoff', 'coulomb_cutoff1',
             'coulomb_cutoff2']
PARAMETERS = [
    'Nmin', 'Nmax', 'desolvationSurfaceScalingFactor', 'desolvationPrefactor',
    'desolvationAllowance', 'coulomb_diel', 'COO_HIS_exception',
    'OCO_HIS_exception', 'CYS_HIS_exception', 'CYS_CYS_exception',
    'min_ligand_model_pka', 'max_ligand_model_pka',
    'include_H_in_interactions', 'coupling_max_number_of_bonds',
    'min_bond_distance_for_hydrogen_bonds', 'coupling_penalty',
    'shared_determinants', 'common_charge_centre', 'hide_penalised_group',
    'remove_penalised_group', 'max_intrinsic_pka_diff',
    'min_interaction_energy', 'max_free_energy_diff', 'min_swap_pka_shift',
    'min_pka', 'max_pka', 'sidechain_interaction']
STRINGS = ['version', 'output_file_tag', 'ligand_typing', 'pH', 'reference']


class Parameters:
    """PROPKA parameter class."""

    def __init__(self):
        """Initialize parameter class.

        Args:
            parameter_file:  file with parameters
        """
        # TODO - need to define all members explicitly
        self.model_pkas = {}
        self.interaction_matrix = InteractionMatrix("interaction_matrix")
        self.sidechain_cutoffs = None
        # TODO - it would be nice to rename these; they're defined everywhere
        self.COO_HIS_exception = None
        self.OCO_HIS_exception = None
        self.CYS_HIS_exception = None
        self.CYS_CYS_exception = None
        # These functions set up remaining data structures implicitly
        self.set_up_data_structures()

    def parse_line(self, line):
        """Parse parameter file line."""
        # first, remove comments
        comment_pos = line.find('#')
        if comment_pos != -1:
            line = line[:comment_pos]
        # split the line into words
        words = line.split()
        if len(words) == 0:
            return
        # parse the words
        if len(words) == 3 and words[0] in NUMBER_DICTIONARIES:
            self.parse_to_number_dictionary(words)
        elif len(words) == 2 and words[0] in STRING_LISTS:
            self.parse_to_string_list(words)
        elif len(words) == 2 and words[0] in DISTANCES:
            self.parse_distance(words)
        elif len(words) == 2 and words[0] in PARAMETERS:
            self.parse_parameter(words)
        elif len(words) == 2 and words[0] in STRINGS:
            self.parse_string(words)
        elif len(words) > 2 and words[0] in LIST_DICTIONARIES:
            self.parse_to_list_dictionary(words)
        elif words[0] in MATRICES+PAIR_WISE_MATRICES:
            self.parse_to_matrix(words)
        elif len(words) == 3 and words[0] in STRING_DICTIONARIES:
            self.parse_to_string_dictionary(words)

    def parse_to_number_dictionary(self, words):
        """Parse field to number dictionary.

        Args:
            words:  strings to parse.
        """
        dict_ = getattr(self, words[0])
        key = words[1]
        value = words[2]
        dict_[key] = float(value)

    def parse_to_string_dictionary(self, words):
        """Parse field to string dictionary.

        Args:
            words:  strings to parse
        """
        dict_ = getattr(self, words[0])
        key = words[1]
        value = words[2]
        dict_[key] = value

    def parse_to_list_dictionary(self, words):
        """Parse field to list dictionary.

        Args:
            words:  strings to parse.
        """
        dict_ = getattr(self, words[0])
        key = words[1]
        if not key in dict_:
            dict_[key] = []
        for value in words[2:]:
            if isinstance(value, list):
                dict_[key].append([float(x) for x in value])
            dict_[key].append(float(value))

    def parse_to_string_list(self, words):
        """Parse field to string list.

        Args:
            words:  strings to parse
        """
        list_ = getattr(self, words[0])
        value = words[1]
        list_.append(value)

    def parse_to_matrix(self, words):
        """Parse field to matrix.

        Args:
            words:  strings to parse
        """
        matrix = getattr(self, words[0])
        value = tuple(words[1:])
        matrix.add(value)

    def parse_distance(self, words):
        """Parse field to distance.

        Args:
            words:  strings to parse
        """
        value = float(words[1])
        setattr(self, words[0], value)
        value_sq = value*value
        attr = "{0:s}_squared".format(words[0])
        setattr(self, attr, value_sq)

    def parse_parameter(self, words):
        """Parse field to parameters.

        Args:
            words:  strings to parse
        """
        value = float(words[1])
        setattr(self, words[0], value)

    def parse_string(self, words):
        """Parse field to strings.

        Args:
            words:  strings to parse
        """
        setattr(self, words[0], words[1])

    def set_up_data_structures(self):
        """Set up internal data structures.

        TODO - it would be better to make these assignments explicit in
        __init__.
        """
        for key_word in (NUMBER_DICTIONARIES + LIST_DICTIONARIES
                         + STRING_DICTIONARIES):
            setattr(self, key_word, {})
        for key_word in STRING_LISTS:
            setattr(self, key_word, [])
        for key_word in STRINGS:
            setattr(self, key_word, "")
        for key_word in MATRICES:
            matrix = InteractionMatrix(key_word)
            setattr(self, key_word, matrix)
        for key_word in PAIR_WISE_MATRICES:
            matrix = PairwiseMatrix(key_word)
            setattr(self, key_word, matrix)

    def print_interaction_parameters(self):
        """Print interaction parameters."""
        info('--------------- Model pKa values ----------------------')
        for k in self.model_pkas:
            info('{0:>3s} {1:8.2f}'.format(k, self.model_pkas[k]))

        info('')
        info('--------------- Interactions --------------------------')
        agroups = [
            'COO', 'HIS', 'CYS', 'TYR', 'SER', 'N+', 'LYS', 'AMD', 'ARG',
            'TRP', 'ROH', 'CG', 'C2N', 'N30', 'N31', 'N32', 'N33', 'NAR',
            'OCO', 'NP1', 'OH', 'O3', 'CL', 'F', 'NAM', 'N1', 'O2', 'OP', 'SH']
        lgroups = [
            'CG', 'C2N', 'N30', 'N31', 'N32', 'N33', 'NAR', 'OCO', 'NP1',
            'OH', 'O3', 'CL', 'F', 'NAM', 'N1', 'O2', 'OP', 'SH']
        map_ = {
            'CG': ['ARG'], 'C2N': ['ARG'], 'N30': ['N+', 'LYS'],
            'N31': ['N+', 'LYS'], 'N32': ['N+', 'LYS'], 'N33': ['N+', 'LYS'],
            'NAR': ['HIS'], 'OCO': ['COO'], 'OP': [], 'SH': ['CYS'],
            'NP1': [], 'OH': ['ROH'], 'O3': [], 'CL': [], 'F': [],
            'NAM': ['AMD'], 'N1': [], 'O2': []}
        for group1 in agroups:
            for group2 in lgroups:
                fmt = "{grp1:>3s} {grp2:>3s} {mat:1s} {val1:4} {val2:4}"
                interaction = fmt.format(
                    grp1=group1, grp2=group2,
                    mat=self.interaction_matrix[group1][group2],
                    val1=self.sidechain_cutoffs.get_value(group1, group2)[0],
                    val2=self.sidechain_cutoffs.get_value(group1, group2)[1])
                map_interaction = ''
                if group2 in map_:
                    for val in map_[group2]:
                        fmt = "|{grp1:>3s} {grp2:>3s} {mat:1s} {val1:4} {val2:4}"
                        map_interaction += fmt.format(
                            group1, val, self.interaction_matrix[group1][val],
                            self.sidechain_cutoffs.get_value(group1, val)[0],
                            self.sidechain_cutoffs.get_value(group1, val)[1])
                        if (self.interaction_matrix[group1][val]
                                != self.interaction_matrix[group1][group2]):
                            map_interaction += '* '
                        if (self.sidechain_cutoffs.get_value(group1, val)[0]
                                != self.sidechain_cutoffs.get_value(
                                    group1, group2)[0]
                                or self.sidechain_cutoffs.get_value(
                                    group1, val)[1]
                                != self.sidechain_cutoffs.get_value(
                                    group1, group2)[1]):
                            map_interaction += '! '
                        else:
                            map_interaction += '  '
                    if (len(map_[group2]) == 0
                            and (self.sidechain_cutoffs.get_value(
                                group1, group2)[0]
                                 != 3
                                 or self.sidechain_cutoffs.get_value(
                                     group1, group2)[1]
                                 != 4)):
                        map_interaction += '?  '
                info(interaction, map_interaction)
                if group1 == group2:
                    break
            info('-')
        info('--------------- Exceptions ----------------------------')
        info('COO-HIS', self.COO_HIS_exception)
        info('OCO-HIS', self.OCO_HIS_exception)
        info('CYS-HIS', self.CYS_HIS_exception)
        info('CYS-CYS', self.CYS_CYS_exception)

        info('--------------- Mapping -------------------------------')
        info("""
Titratable:
CG  ARG
C2N ARG
N30 N+/LYS
N31 N+/LYS
N32 N+/LYS
N33 N+/LYS
NAR HIS
OCO COO
OP  TYR/SER?
SH  CYS

Non-titratable:
NP1 AMD?
OH  ROH
O3  ?
CL
F
NAM
N1
O2
""")

    def print_interaction_parameters_latex(self):
        """Print interaction parameters in LaTeX format."""
        # TODO - if these lists and dictionaries are the same as above, then
        # should be constants at the level of the module
        agroups = ['COO', 'HIS', 'CYS', 'TYR', 'SER', 'N+', 'LYS', 'AMD',
                   'ARG', 'TRP', 'ROH', 'CG', 'C2N', 'N30', 'N31', 'N32',
                   'N33', 'NAR', 'OCO', 'NP1', 'OH', 'O3', 'CL', 'F', 'NAM',
                   'N1', 'O2', 'OP', 'SH']
        lgroups = ['CG', 'C2N', 'N30', 'N31', 'N32', 'N33', 'NAR', 'OCO',
                   'NP1', 'OH', 'O3', 'CL', 'F', 'NAM', 'N1', 'O2', 'OP',
                   'SH']
        lines = [
            "",
            "\\begin{{longtable}}{{lllll}}",
            ("\\caption{{Ligand interaction parameters. For interactions not "
             "listed, the default value of {0:s} is applied.}}").format(
                 self.sidechain_cutoffs.default),
            "\\label{{tab:ligand_interaction_parameters}}\\\\",
            "\\toprule",
            "Group1 & Group2 & Interaction & c1 &c2 \\\\",
            "\\midrule",
            "\\endfirsthead",
            "",
            "\\multicolumn{{5}}{{l}}{\\emph{{continued from the previous page}}}\\\\",
            "\\toprule",
            "Group1 & Group2 & Interaction & c1 &c2 \\\\",
            "\\midrule",
            "\\endhead",
            "",
            "\\midrule",
            "\\multicolumn{{5}}{{r}}{\\emph{{continued on the next page}}}\\\\",
            "\\endfoot",
            "",
            "\\bottomrule",
            "\\endlastfoot",
            ""]
        str_ = "\n".join(lines)
        for group1 in agroups:
            for group2 in lgroups:
                if self.interaction_matrix[group1][group2] == '-':
                    continue
                if (self.sidechain_cutoffs.get_value(group1, group2)
                        == self.sidechain_cutoffs.default):
                    continue
                fmt = (
                    "{grp1:>3s} & {grp2:>3s} & {mat:1s} & {val1:4} & "
                    "{val2:4}\\\\ \n")
                str_ += fmt.format(
                    group1, group2,
                    self.interaction_matrix[group1][group2],
                    self.sidechain_cutoffs.get_value(group1, group2)[0],
                    self.sidechain_cutoffs.get_value(group1, group2)[1])
                if group1 == group2:
                    break
        str_ += '  \\end{{longtable}}\n'
        info(str_)

    def print_interactions_latex(self):
        """Print interactions in LaTeX."""
        # TODO - are these the same lists as above? Convert to module constants.
        agroups = ['COO', 'HIS', 'CYS', 'TYR', 'SER', 'N+', 'LYS', 'AMD',
                   'ARG', 'TRP', 'ROH', 'CG', 'C2N', 'N30', 'N31', 'N32',
                   'N33', 'NAR', 'OCO', 'NP1', 'OH', 'O3', 'CL', 'F', 'NAM',
                   'N1', 'O2', 'OP', 'SH']
        lines = [
            "",
            "\\begin{{longtable}}{{{0:s}}}".format('l'*len(agroups)),
            ("\\caption{{Ligand interaction parameters. For interactions not "
             "listed, the default value of {0:s} is applied.}}").format(
                 str(self.sidechain_cutoffs.default)),
            "\\label{{tab:ligand_interaction_parameters}}\\\\",
            "\\toprule",
            "Group1 & Group2 & Interaction & c1 &c2 \\\\",
            "\\midrule",
            "\\endfirsthead",
            "",
            "\\multicolumn{{5}}{{l}}{\\emph{{continued from the previous page}}}\\\\",
            "\\toprule",
            "Group1 & Group2 & Interaction & c1 &c2 \\\\",
            "\\midrule",
            "\\endhead",
            "",
            "\\midrule",
            "\\multicolumn{{5}}{{r}}{\\emph{{continued on the next page}}}\\\\",
            "\\endfoot",
            "",
            "\\bottomrule",
            "\\endlastfoot",
            ""
        ]
        str_ = "\n".join(lines)
        for group1 in agroups:
            for group2 in agroups:
                fmt = '{g1:>3s} & {g2:>3s} & {mat:1s} & {val1:>4s} & {val2:>4s}\\\\ \n'
                str_ += fmt.format(
                    group1, group2, self.interaction_matrix[group1][group2],
                    str(self.sidechain_cutoffs.get_value(group1, group2)[0]),
                    str(self.sidechain_cutoffs.get_value(group1, group2)[1]))
                if group1 == group2:
                    break
        str_ += '  \\end{{longtable}}\n'
        info(str_)


class InteractionMatrix:
    """Interaction matrix class."""

    def __init__(self, name):
        """Initialize with name of matrix.

        Args:
            name:  name of interaction matrix
        """
        self.name = name
        self.value = None
        self.ordered_keys = []
        self.dictionary = {}

    def add(self, words):
        """Add values to matrix.

        Args:
            words:  values to add
        """
        new_group = words[0]
        self.ordered_keys.append(new_group)
        if not new_group in self.dictionary.keys():
            self.dictionary[new_group] = {}
        for i, group in enumerate(self.ordered_keys):
            if len(words) > i+1:
                try:
                    self.value = float(words[i+1])
                except ValueError:
                    self.value = words[i+1]
                self.dictionary[group][new_group] = self.value
                self.dictionary[new_group][group] = self.value

    def get_value(self, item1, item2):
        """Get specific matrix value.

        Args:
            item1:  matrix row index
            item2:  matrix column index
        Returns:
            matrix value or None
        """
        try:
            return self.dictionary[item1][item2]
        except KeyError:
            return None

    def __getitem__(self, group):
        """Get specific group from matrix.

        Args:
            group:  group to get
        """
        if group not in self.dictionary.keys():
            str_ = '{0:s} not found in interaction matrix {1:s}'.format(
                group, self.name)
            raise KeyError(str_)
        return self.dictionary[group]

    def keys(self):
        """Get keys from matrix.

        Returns:
            dictionary key list
        """
        return self.dictionary.keys()

    def __str__(self):
        str_ = '      '
        for key in self.ordered_keys:
            str_ += '{0:>3s} '.format(key)
        str_ += '\n'
        for key1 in self.ordered_keys:
            str_ += '{0:>3s} '.format(key1)
            for key2 in self.ordered_keys:
                str_ += '{0:>3s} '.format(self[key1][key2])
            str_ += '\n'
        return str_


class PairwiseMatrix:
    """Pairwise interaction matrix class."""

    def __init__(self, name):
        """Initialize pairwise matrix.

        Args:
            name:  name of pairwise interaction
        """
        self.name = name
        self.dictionary = {}
        self.default = [0.0, 0.0]

    def add(self, words):
        """Add information to the matrix.

        TODO - this function unnecessarily bundles arguments into a tuple

        Args:
            words:  tuple with assignment information and value
        """
        # assign the default value
        if len(words) == 3 and words[0] == 'default':
            self.default = [float(words[1]), float(words[2])]
            return
        # assign non-default values
        group1 = words[0]
        group2 = words[1]
        value = [float(words[2]), float(words[3])]
        self.insert(group1, group2, value)
        self.insert(group2, group1, value)

    def insert(self, key1, key2, value):
        """Insert value into matrix.

        Args:
            key1:  first matrix key (row)
            key2:  second matrix key (column)
            value:  value to insert
        """
        if key1 in self.dictionary and key2 in self.dictionary[key1]:
            if key1 != key2:
                str_ = (
                    'Parameter value for {0:s}, {1:s} defined more '
                    'than once'.format(key1, key2))
                warning(str_)
        if not key1 in self.dictionary:
            self.dictionary[key1] = {}
        self.dictionary[key1][key2] = value

    def get_value(self, item1, item2):
        """Get specified value from matrix.

        Args:
            item1:  row index
            item2:  column index
        Returns:
            matrix value (or default)
        """
        try:
            return self.dictionary[item1][item2]
        except KeyError:
            return self.default

    def __getitem__(self, group):
        """Get item from matrix corresponding to specific group.

        Args:
            group:  group to retrieve
        Returns:
            matrix information
        """
        if group not in self.dictionary.keys():
            str_ = '{0:s} not found in interaction matrix {1:s}'.format(
                group, self.name)
            raise KeyError(str_)
        return self.dictionary[group]

    def keys(self):
        """Get keys from matrix.

        Returns:
            dictionary key list
        """
        return self.dictionary.keys()

    def __str__(self):
        str_ = ''
        for key1 in self.keys():
            for key2 in self[key1].keys():
                str_ += '{0:s} {1:s} {2:s}\n'.format(
                    key1, key2, self[key1][key2])
        return str_



"""PROPKA representation of bonds."""
import math
import json
import pkg_resources
import propka.calculations
# TODO - replace the info/warning imports with logging functionality
from propka.lib import info


# TODO - should these constants be defined higher up in the module?
# TODO - I don't know what some of these constants mean
DISULFIDE_DISTANCE = 2.5
FLUORIDE_DISTANCE = 1.7
HYDROGEN_DISTANCE = 1.5
DEFAULT_DISTANCE = 2.0
BOX_SIZE = 2.5
POS_MAX = 1e6


class BondMaker:
    """Makes bonds?

    TODO - the documentation for this class needs to be improved.
    """
    def __init__(self):
        # predefined bonding distances
        self.distances = {'S-S' : DISULFIDE_DISTANCE,
                          'F-F' : FLUORIDE_DISTANCE}
        self.distances_squared = {}
        for key in self.distances:
            self.distances_squared[key] = (
                self.distances[key] * self.distances[key])
        h_dist = HYDROGEN_DISTANCE
        self.default_dist = DEFAULT_DISTANCE
        self.h_dist_squared = h_dist * h_dist
        self.default_dist_squared = self.default_dist * self.default_dist
        distances = (
            list(self.distances_squared.values())
            + [self.default_dist_squared])
        self.max_sq_distance = max(distances)
        # protein bonding data
        self.data_file_name = (
            pkg_resources.resource_filename(__name__, 'protein_bonds.json'))
        with open(self.data_file_name, 'rt') as json_file:
            self.protein_bonds = json.load(json_file)
        self.intra_residue_backbone_bonds = {'N': ['CA'], 'CA': ['N', 'C'],
                                             'C': ['CA', 'O'], 'O': ['C']}
        self.num_pi_elec_bonds_backbone = {'C': 1, 'O': 1}
        self.num_pi_elec_conj_bonds_backbone = {'N': 1}
        self.num_pi_elec_bonds_sidechains = {'ARG-CZ' : 1, 'ARG-NH1': 1,
                                             'ASN-OD1': 1, 'ASN-CG' : 1,
                                             'ASP-OD1': 1, 'ASP-CG' : 1,
                                             'GLU-OE1': 1, 'GLU-CD' : 1,
                                             'GLN-OE1': 1, 'GLN-CD' : 1,
                                             'HIS-CG' : 1, 'HIS-CD2': 1,
                                             'HIS-ND1': 1, 'HIS-CE1': 1,
                                             'PHE-CG' : 1, 'PHE-CD1': 1,
                                             'PHE-CE1': 1, 'PHE-CZ' : 1,
                                             'PHE-CE2': 1, 'PHE-CD2': 1,
                                             'TRP-CG' : 1, 'TRP-CD1': 1,
                                             'TRP-CE2': 1, 'TRP-CD2': 1,
                                             'TRP-CE3': 1, 'TRP-CZ3': 1,
                                             'TRP-CH2': 1, 'TRP-CZ2': 1,
                                             'TYR-CG' : 1, 'TYR-CD1': 1,
                                             'TYR-CE1': 1, 'TYR-CZ' : 1,
                                             'TYR-CE2': 1, 'TYR-CD2': 1}
        self.num_pi_elec_conj_bonds_sidechains = {'ARG-NE': 1, 'ARG-NH2': 1,
                                                  'ASN-ND2': 1, 'GLN-NE2': 1,
                                                  'HIS-NE2': 1, 'TRP-NE1': 1}
        self.num_pi_elec_bonds_ligands = {'C.ar': 1, 'N.pl3': 0, 'C.2': 1,
                                          'O.2': 1, 'O.co2': 1, 'N.ar': 1,
                                          'C.1': 2, 'N.1': 2}
        self.num_pi_elec_conj_bonds_ligands = {'N.am': 1, 'N.pl3': 1}
        self.backbone_atoms = list(self.intra_residue_backbone_bonds.keys())
        self.terminal_oxygen_names = ['OXT', 'O\'\'']
        self.boxes = {}
        self.num_box_x = None
        self.num_box_y = None
        self.num_box_z = None

    def find_bonds_for_protein(self, protein):
        """Bonds proteins based on the way atoms normally bond.

        Args:
            protein:  the protein to search for bonds
        """
        info('++++ Side chains ++++')
        # side chains
        for chain in protein.chains:
            for residue in chain.residues:
                if residue.res_name.replace(' ', '') not in ['N+', 'C-']:
                    self.find_bonds_for_side_chain(residue.atoms)
        info('++++ Backbones ++++')
        # backbone
        last_residues = []
        for chain in protein.chains:
            for i in range(1, len(chain.residues)):
                if (chain.residues[i-1].res_name.replace(' ', '')
                        not in ['N+', 'C-']):
                    if (chain.residues[i].res_name.replace(' ', '')
                            not in ['N+', 'C-']):
                        self.connect_backbone(chain.residues[i-1],
                                              chain.residues[i])
                        last_residues.append(chain.residues[i])
        info('++++ terminal oxygen ++++')
        # terminal OXT
        for last_residue in last_residues:
            self.find_bonds_for_terminal_oxygen(last_residue)
        info('++++ cysteines ++++')
        # Cysteines
        for chain in protein.chains:
            for i in range(0, len(chain.residues)):
                if chain.residues[i].res_name == 'CYS':
                    for j in range(0, len(chain.residues)):
                        if chain.residues[j].res_name == 'CYS' and j != i:
                            self.check_for_cysteine_bonds(chain.residues[i],
                                                          chain.residues[j])

    def check_for_cysteine_bonds(self, cys1, cys2):
        """Looks for potential bonds between two cysteines.

        Args:
            cys1:  one of the cysteines to check
            cys1:  one of the cysteines to check
        """
        for atom1 in cys1.atoms:
            if atom1.name == 'SG':
                for atom2 in cys2.atoms:
                    if atom2.name == 'SG':
                        dist = propka.calculations.squared_distance(atom1,
                                                                    atom2)
                        # TODO - is SS_dist_squared an attribute of this class?
                        if dist < self.SS_dist_squared:
                            self.make_bond(atom1, atom2)

    def find_bonds_for_terminal_oxygen(self, residue):
        """Look for bonds for terminal oxygen.

        Args:
            residue - test residue
        """
        for atom1 in residue.atoms:
            if atom1.name in self.terminal_oxygen_names:
                for atom2 in residue.atoms:
                    if atom2.name == 'C':
                        self.make_bond(atom1, atom2)

    def connect_backbone(self, residue1, residue2):
        """Sets up bonds in the backbone

        Args:
            residue1:  first residue to connect
            residue2:  second residue to connect
        """
        self.find_bonds_for_residue_backbone(residue1)
        self.find_bonds_for_residue_backbone(residue2)
        for atom1 in residue1.atoms:
            if atom1.name == 'C':
                for atom2 in residue2.atoms:
                    if atom2.name == 'N':
                        if (propka.calculations.squared_distance(atom1, atom2)
                                < self.default_dist_squared):
                            self.make_bond(atom1, atom2)

    def find_bonds_for_residue_backbone(self, residue):
        """Find bonds for this residue's backbone.

        Args:
            residue:  reside to search for backbone bonds.
        """
        for atom1 in residue.atoms:
            if atom1.name in list(self.num_pi_elec_bonds_backbone.keys()):
                atom1.num_pi_elec_2_3_bonds = (
                    self.num_pi_elec_bonds_backbone[atom1.name])
            if atom1.name in (
                    list(self.num_pi_elec_conj_bonds_backbone.keys())
                    and len(atom1.bonded_atoms) > 1): # avoid N-term
                atom1.num_pi_elec_conj_2_3_bonds = (
                    self.num_pi_elec_conj_bonds_backbone[atom1.name])

            if atom1.name in self.backbone_atoms:
                for atom2 in residue.atoms:
                    if atom2.name in (
                            self.intra_residue_backbone_bonds[atom1.name]):
                        self.make_bond(atom1, atom2)

    def find_bonds_for_side_chain(self, atoms):
        """Finds bonds for a side chain.

        Args:
            atoms:  list of atoms to check for bonds
        """
        for atom1 in atoms:
            key = '{0:s}-{1:s}'.format(atom1.res_name, atom1.name)
            if key in list(self.num_pi_elec_bonds_sidechains.keys()):
                atom1.num_pi_elec_2_3_bonds = (
                    self.num_pi_elec_bonds_sidechains[key])
            if key in list(self.num_pi_elec_conj_bonds_sidechains.keys()):
                atom1.num_pi_elec_conj_2_3_bonds = (
                    self.num_pi_elec_conj_bonds_sidechains[key])
            if not atom1.name in self.backbone_atoms:
                if not atom1.name in self.terminal_oxygen_names:
                    for atom2 in atoms:
                        if atom2.name in (
                                self
                                .protein_bonds[atom1.res_name][atom1.name]):
                            self.make_bond(atom1, atom2)

    def find_bonds_for_ligand(self, ligand):
        """Finds bonds for all atoms in the ligand molecule

        Args:
            ligand:  ligand molecule to search for bonds
        """
        # identify bonding atoms
        self.find_bonds_for_atoms(ligand.atoms)
        self.add_pi_electron_table_info(ligand.atoms)

    def add_pi_electron_table_info(self, atoms):
        """Add table information on pi electrons

        Args:
            atoms:  list of atoms for pi electron table information checking
        """
        # apply table information on pi-electrons
        for atom in atoms:
            # for ligands
            if atom.type == 'hetatm':
                if atom.sybyl_type in self.num_pi_elec_bonds_ligands.keys():
                    atom.num_pi_elec_2_3_bonds = (
                        self.num_pi_elec_bonds_ligands[atom.sybyl_type])
                if atom.sybyl_type in (
                        self.num_pi_elec_conj_bonds_ligands.keys()):
                    atom.num_pi_elec_conj_2_3_bonds = (
                        self.num_pi_elec_conj_bonds_ligands[atom.sybyl_type])
            # for protein
            if atom.type == 'atom':
                key = '{0:s}-{1:s}'.format(atom.res_name, atom.name)
                if key in list(self.num_pi_elec_bonds_sidechains.keys()):
                    atom.num_pi_elec_2_3_bonds = (
                        self.num_pi_elec_bonds_sidechains[key])
                if key in list(self.num_pi_elec_conj_bonds_sidechains.keys()):
                    atom.num_pi_elec_conj_2_3_bonds = (
                        self.num_pi_elec_conj_bonds_sidechains[key])
                if atom.name in list(self.num_pi_elec_bonds_backbone.keys()):
                    atom.num_pi_elec_2_3_bonds = (
                        self.num_pi_elec_bonds_backbone[atom.name])
                if atom.name in list(
                        self.num_pi_elec_conj_bonds_backbone.keys()) and (
                            len(atom.bonded_atoms) > 1):
                    # last part to avoid including N-term
                    atom.num_pi_elec_conj_2_3_bonds = (
                        self.num_pi_elec_conj_bonds_backbone[atom.name])

    def find_bonds_for_protein_by_distance(self, molecule):
        """Finds bonds for all atoms in the molecule.

        Args:
            molecule:  molecule in which to find bonds.
        Returns:
            list of atoms
        """
        #self.find_bonds_for_protein(molecule)
        atoms = []
        for chain in molecule.chains:
            for residue in chain.residues:
                if residue.res_name.replace(' ', '') not in ['N+', 'C-']:
                    for atom in residue.atoms:
                        atoms.append(atom)
        self.find_bonds_for_atoms_using_boxes(atoms)
        return atoms

    def find_bonds_for_atoms(self, atoms):
        """Finds all bonds for a list of atoms

        Args:
            atoms:  list of atoms in which to find bonds.
        """
        no_atoms = len(atoms)
        for i in range(no_atoms):
            for j in range(i+1, no_atoms):
                if atoms[i] in atoms[j].bonded_atoms:
                    continue
                if self.check_distance(atoms[i], atoms[j]):
                    self.make_bond(atoms[i], atoms[j])
                    # di-sulphide bonds
                    if atoms[i].element == 'S' and atoms[j].element == 'S':
                        atoms[i].cysteine_bridge = True
                        atoms[j].cysteine_bridge = True

    def check_distance(self, atom1, atom2):
        """Check distance between two atoms

        Args:
            atom1:  first atom for distance check
            atom2:  second atom for distance check
        Returns:
            True if within distance, False otherwise
        """
        sq_dist = propka.calculations.squared_distance(atom1, atom2)
        if sq_dist > self.max_sq_distance:
            return False
        key = '{0:s}-{1:s}'.format(atom1.element, atom2.element)
        h_count = key.count('H')
        if sq_dist < self.h_dist_squared and h_count == 1:
            return True
        if sq_dist < self.default_dist_squared and h_count == 0:
            return True
        if key in self.distances_squared.keys():
            if sq_dist < self.distances_squared[key]:
                return True
        return False

    def find_bonds_for_molecules_using_boxes(self, molecules):
        """ Finds all bonds for a molecular container.

        Args:
            molecules:  list of molecules for finding bonds.
        """
        for name in molecules.conformation_names:
            self.find_bonds_for_atoms_using_boxes(
                molecules.conformations[name].atoms)

    def add_pi_electron_information(self, molecules):
        """Add pi electron information to a molecule.

        Args:
            molecules:  list of molecules for adding pi electron information.
        """
        for name in molecules.conformation_names:
            self.add_pi_electron_table_info(
                molecules.conformations[name].atoms)

    def find_bonds_for_atoms_using_boxes(self, atoms):
        """Finds all bonds for a list of atoms.

        Args:
            atoms:  list of atoms for finding bonds
        """
        box_size = BOX_SIZE
        xmin = POS_MAX
        xmax = -1.0 * POS_MAX
        ymin = POS_MAX
        ymax = -1.0 * POS_MAX
        zmin = POS_MAX
        zmax = -1.0 * POS_MAX
        for atom in atoms:
            if atom.x > xmax:
                xmax = atom.x
            if atom.y > ymax:
                ymax = atom.y
            if atom.z > zmax:
                zmax = atom.z
            if atom.x < xmin:
                xmin = atom.x
            if atom.y < ymin:
                ymin = atom.y
            if atom.z < zmin:
                zmin = atom.z
        xlen = xmax - xmin
        ylen = ymax - ymin
        zlen = zmax - zmin
        self.num_box_x = max(1, int(math.ceil(xlen/box_size)))
        self.num_box_y = max(1, int(math.ceil(ylen/box_size)))
        self.num_box_z = max(1, int(math.ceil(zlen/box_size)))
        self.boxes = {}
        for x in range(self.num_box_x):
            for y in range(self.num_box_y):
                for z in range(self.num_box_z):
                    self.boxes[(x, y, z)] = []
        for atom in atoms:
            x = math.floor((atom.x - xmin)/box_size)
            y = math.floor((atom.y - ymin)/box_size)
            z = math.floor((atom.z - zmin)/box_size)
            self.put_atom_in_box(x, y, z, atom)
        for value in self.boxes.values():
            self.find_bonds_for_atoms(value)

    def put_atom_in_box(self, x, y, z, atom):
        """Put an atom in a box.

        Args:
            x:  box x-coordinates
            y:  box y-coordinates
            z:  box z-coordinates
            atom:  the atom to place in a box
        """
        for box_x in [x, x+1]:
            for box_y in [y, y+1]:
                for box_z in [z, z+1]:
                    key = (box_x, box_y, box_z)
                    try:
                        self.boxes[key].append(atom)
                    except KeyError:
                        pass

    @staticmethod
    def has_bond(atom1, atom2):
        """Look for bond between two atoms.

        Args:
            atom1:  first atom to check
            atom2:  second atom to check
        Returns:
            True if there is a bond between atoms
        """
        if atom1 in atom2.bonded_atoms or atom2 in atom1.bonded_atoms:
            return True
        return False

    @staticmethod
    def make_bond(atom1, atom2):
        """Makes a bond between atom1 and atom2

        Args:
            atom1:  first atom to bond
            atom2:  second atom to bond
        """
        if atom1 == atom2:
            return
        if not atom1 in atom2.bonded_atoms:
            atom2.bonded_atoms.append(atom1)
        if not atom2 in atom1.bonded_atoms:
            atom1.bonded_atoms.append(atom2)

    def generate_protein_bond_dictionary(self, atoms):
        """Generate dictionary of protein bonds.

        Args:
            atoms:  list of atoms for bonding
        """
        for atom in atoms:
            for bonded_atom in atom.bonded_atoms:
                resi_i = atom.res_name
                name_i = atom.name
                resi_j = bonded_atom.res_name
                name_j = bonded_atom.name
                if not name_i in (
                        self.backbone_atoms
                        or not name_j in self.backbone_atoms):
                    if not name_i in (
                            self.terminal_oxygen_names
                            and not name_j in self.terminal_oxygen_names):
                        if not resi_i in list(self.protein_bonds.keys()):
                            self.protein_bonds[resi_i] = {}
                        if not name_i in self.protein_bonds[resi_i]:
                            self.protein_bonds[resi_i][name_i] = []
                        if not name_j in self.protein_bonds[resi_i][name_i]:
                            self.protein_bonds[resi_i][name_i].append(name_j)
                        if not resi_j in list(self.protein_bonds.keys()):
                            self.protein_bonds[resi_j] = {}
                        if not name_j in self.protein_bonds[resi_j]:
                            self.protein_bonds[resi_j][name_j] = []
                        if not name_i in self.protein_bonds[resi_j][name_j]:
                            self.protein_bonds[resi_j][name_j].append(name_i)



"""PROPKA 3

See https://github.com/jensengroup/propka-3.1 for more information.

Please cite these PROPKA references in publications:

* Sondergaard, Chresten R., Mats HM Olsson, Michal Rostkowski, and Jan H. Jensen.
"Improved Treatment of Ligands and Coupling Effects in Empirical Calculation and
Rationalization of pKa Values." Journal of Chemical Theory and Computation 7,
no. 7 (2011): 2284-2295.

* Olsson, Mats HM, Chresten R. Sondergaard, Michal Rostkowski, and Jan H. Jensen.
"PROPKA3: consistent treatment of internal and surface residues in empirical pKa
predictions." Journal of Chemical Theory and Computation 7, no. 2 (2011): 525-537.
"""
__all__ = ["atom", "bonds", "calculations", "conformation_container",
           "coupled_groups", "determinant", "determinants", "group",
           "hybrid36", "iterative", "input", "lib", "ligand_pka_values",
           "ligand", "molecular_container", "output", "parameters",
           "protonate", "run", "vector_algebra", "version"]



"""Energy calculations."""
import math
from propka.lib import warning
from propka.calculations import squared_distance, get_smallest_distance


# TODO - I have no idea what these constants mean so I labeled them "UNK_"
UNK_MIN_DISTANCE = 2.75
MIN_DISTANCE_4TH = math.pow(UNK_MIN_DISTANCE, 4)
UNK_DIELECTRIC1 = 160
UNK_DIELECTRIC2 = 30
UNK_PKA_SCALING1 = 244.12
UNK_BACKBONE_DISTANCE1 = 6.0
UNK_BACKBONE_DISTANCE2 = 3.0
UNK_FANGLE_MIN = 0.001
UNK_PKA_SCALING2 = 0.8
COMBINED_NUM_BURIED_MAX = 900
SEPARATE_NUM_BURIED_MAX = 400


def radial_volume_desolvation(parameters, group):
    """Calculate desolvation terms for group.

    Args:
        parameters:  parameters for desolvation calculation
        group:  group of atoms for calculation
    """
    all_atoms = group.atom.conformation_container.get_non_hydrogen_atoms()
    volume = 0.0
    group.num_volume = 0
    min_dist_4th = MIN_DISTANCE_4TH
    for atom in all_atoms:
        # ignore atoms in the same residue
        if (atom.res_num == group.atom.res_num
                and atom.chain_id == group.atom.chain_id):
            continue
        sq_dist = squared_distance(group, atom)
        # desolvation
        if sq_dist < parameters.desolv_cutoff_squared:
            # use a default relative volume of 1.0 if the volume of the element
            # is not found in parameters
            # insert check for methyl groups
            if atom.element == 'C' and atom.name not in ['CA', 'C']:
                dvol = parameters.VanDerWaalsVolume['C4']
            else:
                dvol = parameters.VanDerWaalsVolume.get(atom.element, 1.0)
            dv_inc = dvol/max(min_dist_4th, sq_dist*sq_dist)
            volume += dv_inc
        # buried
        if sq_dist < parameters.buried_cutoff_squared:
            group.num_volume += 1
    group.buried = calculate_weight(parameters, group.num_volume)
    scale_factor = calculate_scale_factor(parameters, group.buried)
    volume_after_allowance = max(0.00, volume-parameters.desolvationAllowance)
    group.energy_volume = (
        group.charge * parameters.desolvationPrefactor
        * volume_after_allowance * scale_factor)


def calculate_scale_factor(parameters, weight):
    """Calculate desolvation scaling factor.

    Args:
        parameters:  parameters for desolvation calculation
        weight:  weight for scaling factor
    Returns:
        scaling factor
    """
    scale_factor = 1.0 - (1.0 - parameters.desolvationSurfaceScalingFactor)*(1.0 - weight)
    return scale_factor


def calculate_weight(parameters, num_volume):
    """Calculate the atom-based desolvation weight.

    TODO - figure out why a similar function exists in version.py

    Args:
        parameters:  parameters for desolvation calculation
        num_volume:  number of heavy atoms within desolvation calculation volume
    Returns:
        desolvation weight
    """
    weight = (
        float(num_volume - parameters.Nmin)
        / float(parameters.Nmax - parameters.Nmin))
    weight = min(1.0, weight)
    weight = max(0.0, weight)
    return weight


def calculate_pair_weight(parameters, num_volume1, num_volume2):
    """Calculate the atom-pair based desolvation weight.

    Args:
        num_volume1:  number of heavy atoms within first desolvation volume
        num_volume2:  number of heavy atoms within second desolvation volume
    Returns:
        desolvation weight
    """
    num_volume = num_volume1 + num_volume2
    num_min = 2*parameters.Nmin
    num_max = 2*parameters.Nmax
    weight = float(num_volume - num_min)/float(num_max - num_min)
    weight = min(1.0, weight)
    weight = max(0.0, weight)
    return weight


def hydrogen_bond_energy(dist, dpka_max, cutoffs, f_angle=1.0):
    """Calculate hydrogen-bond interaction pKa shift.

    Args:
        dist:  distance for hydrogen bond
        dpka_max:  maximum pKa value shift
        cutoffs:  array with max and min distance values
        f_angle:  angle scaling factor
    Returns:
        pKa shift value
    """
    if dist < cutoffs[0]:
        value = 1.00
    elif dist > cutoffs[1]:
        value = 0.00
    else:
        value = 1.0 - (dist - cutoffs[0])/(cutoffs[1] - cutoffs[0])
    dpka = dpka_max*value*f_angle
    return abs(dpka)


def angle_distance_factors(atom1=None, atom2=None, atom3=None, center=None):
    """Calculate distance and angle factors for three atoms for backbone interactions.

    NOTE - you need to use atom1 to be the e.g. ASP atom if distance is reset at
    return: [O1 -- H2-N3].

    Also generalized to be able to be used for residue 'centers' for C=O COO interactions.

    Args:
        atom1:  first atom for calculation (could be None)
        atom2:  second atom for calculation
        atom3:  third atom for calculation
        center:  center point between atoms 1 and 2
    Returns
        [distance factor between atoms 1 and 2,
         angle factor,
         distance factor between atoms 2 and 3]
    """
    # The angle factor
    #
    #  ---closest_atom1/2
    #                     .
    #                      .
    #                       the_hydrogen---closest_atom2/1---
    dx_32 = atom2.x - atom3.x
    dy_32 = atom2.y - atom3.y
    dz_32 = atom2.z - atom3.z
    dist_23 = math.sqrt(dx_32 * dx_32 + dy_32 * dy_32 + dz_32 * dz_32)
    dx_32 = dx_32/dist_23
    dy_32 = dy_32/dist_23
    dz_32 = dz_32/dist_23
    if atom1 is None:
        dx_21 = center[0] - atom2.x
        dy_21 = center[1] - atom2.y
        dz_21 = center[2] - atom2.z
    else:
        dx_21 = atom1.x - atom2.x
        dy_21 = atom1.y - atom2.y
        dz_21 = atom1.z - atom2.z
    dist_12 = math.sqrt(dx_21 * dx_21 + dy_21 * dy_21 + dz_21 * dz_21)
    dx_21 = dx_21/dist_12
    dy_21 = dy_21/dist_12
    dz_21 = dz_21/dist_12
    f_angle = dx_21*dx_32 + dy_21*dy_32 + dz_21*dz_32
    return dist_12, f_angle, dist_23


def hydrogen_bond_interaction(group1, group2, version):
    """Calculate energy for hydrogen bond interactions between two groups.

    Args:
        group1:  first interacting group
        group2:  second interacting group
        version:  an object that contains version-specific parameters
    Returns:
        hydrogen bond interaction energy
    """
    # find the smallest distance between interacting atoms
    atoms1 = group1.get_interaction_atoms(group2)
    atoms2 = group2.get_interaction_atoms(group1)
    [closest_atom1, dist, closest_atom2] = get_smallest_distance(atoms1, atoms2)
    if None in [closest_atom1, closest_atom2]:
        warning(
            'Side chain interaction failed for {0:s} and {1:s}'.format(
                group1.label, group2.label))
        return None
    # get the parameters
    [dpka_max, cutoff] = version.get_hydrogen_bond_parameters(closest_atom1,
                                                              closest_atom2)
    if (dpka_max is None) or (None in cutoff):
        return None
    # check that the closest atoms are close enough
    if dist >= cutoff[1]:
        return None
    # check that bond distance criteria is met
    min_hbond_dist = version.parameters.min_bond_distance_for_hydrogen_bonds
    if group1.atom.is_atom_within_bond_distance(group2.atom, min_hbond_dist, 1):
        return None
    # set angle factor
    f_angle = 1.0
    if group2.type in version.parameters.angular_dependent_sidechain_interactions:
        if closest_atom2.element == 'H':
            heavy_atom = closest_atom2.bonded_atoms[0]
            hydrogen = closest_atom2
            dist, f_angle, _ = angle_distance_factors(closest_atom1, hydrogen,
                                                      heavy_atom)
        else:
            # Either the structure is corrupt (no hydrogen), or the heavy atom
            # is closer to the titratable atom than the hydrogen. In either
            # case, we set the angle factor to 0
            f_angle = 0.0
    elif group1.type in version.parameters.angular_dependent_sidechain_interactions:
        if closest_atom1.element == 'H':
            heavy_atom = closest_atom1.bonded_atoms[0]
            hydrogen = closest_atom1
            dist, f_angle, _ = angle_distance_factors(closest_atom2, hydrogen,
                                                      heavy_atom)
        else:
            # Either the structure is corrupt (no hydrogen), or the heavy atom
            # is closer to the titratable atom than the hydrogen. In either
            # case, we set the angle factor to 0
            f_angle = 0.0
    weight = version.calculate_pair_weight(group1.num_volume, group2.num_volume)
    exception, value = version.check_exceptions(group1, group2)
    if exception:
        # Do nothing, value should have been assigned.
        pass
    else:
        value = version.calculate_side_chain_energy(
            dist, dpka_max, cutoff, weight, f_angle)
    return value


def electrostatic_interaction(group1, group2, dist, version):
    """Calculate electrostatic interaction betwee two groups.

    Args:
        group1:  first interacting group
        group2:  second interacting group
        dist:  distance between groups
        version:  version-specific object with parameters and functions
    Returns:
        electrostatic interaction energy or None (if no interaction is appropriate)
    """
    # check if we should do coulomb interaction at all
    if not version.check_coulomb_pair(group1, group2, dist):
        return None
    weight = version.calculate_pair_weight(group1.num_volume, group2.num_volume)
    value = version.calculate_coulomb_energy(dist, weight)
    return value


def check_coulomb_pair(parameters, group1, group2, dist):
    """Checks if this Coulomb interaction should be done.

    NOTE - this is a propka2.0 hack
    TODO - figure out why a similar function exists in version.py

    Args:
        parameters:  parameters for Coulomb calculations
        group1:  first interacting group
        group2:  second interacting group
        dist:  distance between groups
    Returns:
        Boolean
    """
    num_volume = group1.num_volume + group2.num_volume
    do_coulomb = True
    # check if both groups are titratable (ions are taken care of in
    # determinants::set_ion_determinants)
    if not (group1.titratable and group2.titratable):
        do_coulomb = False
    # check if the distance is not too big
    if dist > parameters.coulomb_cutoff2:
        do_coulomb = False
    # check that desolvation is ok
    if num_volume < parameters.Nmin:
        do_coulomb = False
    return do_coulomb


def coulomb_energy(dist, weight, parameters):
    """Calculates the Coulomb interaction pKa shift based on Coulomb's law.

    Args:
        dist:  distance for electrostatic interaction
        weight:  scaling of dielectric constant
        parameters:  parameter object for calculation
    Returns:
        pKa shift
    """
    diel = UNK_DIELECTRIC1 - (UNK_DIELECTRIC1 - UNK_DIELECTRIC2)*weight
    dist = max(dist, parameters.coulomb_cutoff1)
    scale = (
        (dist - parameters.coulomb_cutoff2)
        / (parameters.coulomb_cutoff1 - parameters.coulomb_cutoff2))
    scale = max(0.0, scale)
    scale = min(1.0, scale)
    dpka = UNK_PKA_SCALING1/(diel*dist)*scale
    return abs(dpka)


def backbone_reorganization(_, conformation):
    """Perform calculations related to backbone reorganizations.

    NOTE - this was described in the code as "adding test stuff"
    NOTE - this function does not appear to be used
    TODO - figure out why a similar function exists in version.py

    Args:
        _:  not used
        conformation:  specific molecule conformation
    """
    titratable_groups = conformation.get_backbone_reorganisation_groups()
    bbc_groups = conformation.get_backbone_co_groups()

    for titratable_group in titratable_groups:
        weight = titratable_group.buried
        dpka = 0.00
        for bbc_group in bbc_groups:
            center = [titratable_group.x, titratable_group.y, titratable_group.z]
            atom2 = bbc_group.get_interaction_atoms(titratable_group)[0]
            dist, f_angle, _ = angle_distance_factors(atom2=atom2,
                                                      atom3=bbc_group.atom,
                                                      center=center)
            if dist < UNK_BACKBONE_DISTANCE1 and f_angle > UNK_FANGLE_MIN:
                value = (
                    1.0 - (dist-UNK_BACKBONE_DISTANCE2)
                    / (UNK_BACKBONE_DISTANCE1-UNK_BACKBONE_DISTANCE2))
                dpka += UNK_PKA_SCALING2 * min(1.0, value)
        titratable_group.energy_local = dpka*weight


def check_exceptions(version, group1, group2):
    """Checks for atypical behavior in interactions between two groups.
    Checks are made based on group type.

    TODO - figure out why a similar function exists in version.py

    Args:
        version:  version object
        group1:  first group for check
        group2:  second group for check
    Returns:
        1. Boolean indicating atypical behavior,
        2. value associated with atypical interaction (None if Boolean is False)
    """
    res_type1 = group1.type
    res_type2 = group2.type
    if (res_type1 == "COO") and (res_type2 == "ARG"):
        exception, value = check_coo_arg_exception(group1, group2, version)
    elif (res_type1 == "ARG") and (res_type2 == "COO"):
        exception, value = check_coo_arg_exception(group2, group1, version)
    elif (res_type1 == "COO") and (res_type2 == "COO"):
        exception, value = check_coo_coo_exception(group1, group2, version)
    elif (res_type1 == "CYS") and (res_type2 == "CYS"):
        exception, value = check_cys_cys_exception(group1, group2, version)
    elif ((res_type1 == "COO") and (res_type2 == "HIS")
          or (res_type1 == "HIS") and (res_type2 == "COO")):
        exception, value = check_coo_his_exception(group1, group2, version)
    elif ((res_type1 == "OCO") and (res_type2 == "HIS")
          or (res_type1 == "HIS") and (res_type2 == "OCO")):
        exception, value = check_oco_his_exception(group1, group2, version)
    elif ((res_type1 == "CYS") and (res_type2 == "HIS")
          or (res_type1 == "HIS") and (res_type2 == "CYS")):
        exception, value = check_cys_his_exception(group1, group2, version)
    else:
        # do nothing, no exception for this pair
        exception = False
        value = None
    return exception, value


def check_coo_arg_exception(group_coo, group_arg, version):
    """Check for COO-ARG interaction atypical behavior.

    Uses the two shortest unique distances (involving 2+2 atoms)

    Args:
        group_coo:  COO group
        group_arg:  ARG group
        version:  version object
    Returns:
        1. Boolean indicating atypical behavior,
        2. value associated with atypical interaction (None if Boolean is False)
    """
    exception = True
    value_tot = 0.00
    # needs to be this way since you want to find shortest distance first
    atoms_coo = []
    atoms_coo.extend(group_coo.get_interaction_atoms(group_arg))
    atoms_arg = []
    atoms_arg.extend(group_arg.get_interaction_atoms(group_coo))
    for _ in ["shortest", "runner-up"]:
        # find the closest interaction pair
        [closest_coo_atom, dist, closest_arg_atom] = get_smallest_distance(atoms_coo,
                                                                           atoms_arg)
        [dpka_max, cutoff] = version.get_hydrogen_bond_parameters(closest_coo_atom,
                                                                  closest_arg_atom)
        # calculate and sum up interaction energy
        f_angle = 1.00
        if group_arg.type in version.parameters.angular_dependent_sidechain_interactions:
            atom3 = closest_arg_atom.bonded_atoms[0]
            dist, f_angle, _ = angle_distance_factors(closest_coo_atom,
                                                      closest_arg_atom,
                                                      atom3)
        value = hydrogen_bond_energy(dist, dpka_max, cutoff, f_angle)
        value_tot += value
        # remove closest atoms before we attemp to find the runner-up pair
        atoms_coo.remove(closest_coo_atom)
        atoms_arg.remove(closest_arg_atom)
    return exception, value_tot


def check_coo_coo_exception(group1, group2, version):
    """Check for COO-COO hydrogen-bond atypical interaction behavior.

    Args:
        group1:  first group for check
        group2:  second group for check
        version:  version object
    Returns:
        1. Boolean indicating atypical behavior,
        2. value associated with atypical interaction (None if Boolean is False)
    """
    exception = True
    interact_groups12 = group1.get_interaction_atoms(group2)
    interact_groups21 = group2.get_interaction_atoms(group1)
    [closest_atom1, dist, closest_atom2] = get_smallest_distance(interact_groups12,
                                                                 interact_groups21)
    [dpka_max, cutoff] = version.get_hydrogen_bond_parameters(closest_atom1,
                                                              closest_atom2)
    f_angle = 1.00
    value = hydrogen_bond_energy(dist, dpka_max, cutoff, f_angle)
    weight = calculate_pair_weight(version.parameters, group1.num_volume, group2.num_volume)
    value = value * (1.0 + weight)
    return exception, value


def check_coo_his_exception(group1, group2, version):
    """Check for COO-HIS atypical interaction behavior

    Args:
        group1:  first group for check
        group2:  second group for check
        version:  version object
    Returns:
        1. Boolean indicating atypical behavior,
        2. value associated with atypical interaction (None if Boolean is False)
    """
    exception = False
    if check_buried(group1.num_volume, group2.num_volume):
        exception = True
    return exception, version.parameters.COO_HIS_exception


def check_oco_his_exception(group1, group2, version):
    """Check for OCO-HIS atypical interaction behavior

    Args:
        group1:  first group for check
        group2:  second group for check
        version:  version object
    Returns:
        1. Boolean indicating atypical behavior,
        2. value associated with atypical interaction (None if Boolean is False)
    """
    exception = False
    if check_buried(group1.num_volume, group2.num_volume):
        exception = True
    return exception, version.parameters.OCO_HIS_exception


def check_cys_his_exception(group1, group2, version):
    """Check for CYS-HIS atypical interaction behavior

    Args:
        group1:  first group for check
        group2:  second group for check
        version:  version object
    Returns:
        1. Boolean indicating atypical behavior,
        2. value associated with atypical interaction (None if Boolean is False)
    """
    exception = False
    if check_buried(group1.num_volume, group2.num_volume):
        exception = True
    return exception, version.parameters.CYS_HIS_exception


def check_cys_cys_exception(group1, group2, version):
    """Check for CYS-CYS atypical interaction behavior

    Args:
        group1:  first group for check
        group2:  second group for check
        version:  version object
    Returns:
        1. Boolean indicating atypical behavior,
        2. value associated with atypical interaction (None if Boolean is False)
    """
    exception = False
    if check_buried(group1.num_volume, group2.num_volume):
        exception = True
    return exception, version.parameters.CYS_CYS_exception


def check_buried(num_volume1, num_volume2):
    """Check to see if an interaction is buried

    Args:
        num_volume1:  number of buried heavy atoms in volume 1
        num_volume2:  number of buried heavy atoms in volume 2
    Returns:
        True if interaction is buried, False otherwise
    """
    if ((num_volume1 + num_volume2 <= COMBINED_NUM_BURIED_MAX)
            and (num_volume1 <= SEPARATE_NUM_BURIED_MAX
                 or num_volume2 <= SEPARATE_NUM_BURIED_MAX)):
        return False
    return True



"""Holds the Determinant class

TODO - it is confusing to have both `determinant.py` and `determinants.py`.
Should these be merged?
"""


class Determinant:
    """Determinant class.

    Appears to be a container for storing information and values about
    groups that interact to influence titration states.

    TODO - figure out what this class does.
    """

    def __init__(self, group, value):
        """Initialize the object.

        Args:
            group:  group associated with Determinant object
            value:  value to assign to group
        """
        self.group = group
        self.label = group.label
        self.value = value

    def add(self, value):
        """Increment determinant value.

        Args:
            value:  value to add to determinant
        """
        self.value += value

    def __str__(self):
        return '{0:s}: {1:8.2f}'.format(self.label, self.value)



"""Calculations related to hydrogen placement."""
import math
from propka.lib import info
from propka.protonate import Protonate
from propka.bonds import BondMaker
from propka.atom import Atom



def setup_bonding_and_protonation(molecular_container):
    """Set up bonding and protonation for a molecule.

    Args:
        parameters:  not used
        molecular_container:  molecule container.
    """
    # make bonds
    my_bond_maker = setup_bonding(molecular_container)
    # set up ligand atom names
    set_ligand_atom_names(molecular_container)
    # apply information on pi electrons
    my_bond_maker.add_pi_electron_information(molecular_container)
    # Protonate atoms
    if molecular_container.options.protonate_all:
        protonator = Protonate(verbose=False)
        protonator.protonate(molecular_container)


def setup_bonding(molecular_container):
    """Set up bonding for a molecular container.

    Args:
        molecular_container:  the molecular container in question
    Returns:
        BondMaker object
    """
    my_bond_maker = BondMaker()
    my_bond_maker.find_bonds_for_molecules_using_boxes(molecular_container)
    return my_bond_maker


def setup_bonding_and_protonation_30_style(molecular_container):
    """Set up bonding for a molecular container.

    Args:
        molecular_container:  the molecular container in question
    Returns:
        BondMaker object
    """
    # Protonate atoms
    protonate_30_style(molecular_container)
    # make bonds
    bond_maker = BondMaker()
    bond_maker.find_bonds_for_molecules_using_boxes(molecular_container)
    return bond_maker


def protonate_30_style(molecular_container):
    """Protonate the molecule.

    Args:
        molecular_container:  molecule
    """
    for name in molecular_container.conformation_names:
        info('Now protonating', name)
        # split atom into residues
        curres = -1000000
        residue = []
        o_atom = None
        c_atom = None
        for atom in molecular_container.conformations[name].atoms:
            if atom.res_num != curres:
                curres = atom.res_num
                if len(residue) > 0:
                    #backbone
                    [o_atom, c_atom] = add_backbone_hydrogen(
                        residue, o_atom, c_atom)
                    #arginine
                    if residue[0].res_name == 'ARG':
                        add_arg_hydrogen(residue)
                    #histidine
                    if residue[0].res_name == 'HIS':
                        add_his_hydrogen(residue)
                    #tryptophan
                    if residue[0].res_name == 'TRP':
                        add_trp_hydrogen(residue)
                    #amides
                    if residue[0].res_name in ['GLN', 'ASN']:
                        add_amd_hydrogen(residue)
                    residue = []
            if atom.type == 'atom':
                residue.append(atom)


def set_ligand_atom_names(molecular_container):
    """Set names for ligands in molecular container.

    Args:
        molecular_container:  molecular container for ligand names
    """
    for name in molecular_container.conformation_names:
        molecular_container.conformations[name].set_ligand_atom_names()


def add_arg_hydrogen(residue):
    """Adds Arg hydrogen atoms to residues according to the 'old way'.

    Args:
        residue:  arginine residue to protonate
    Returns:
        list of hydrogen atoms
    """
    #info('Adding arg H',residue)
    for atom in residue:
        if atom.name == "CD":
            cd_atom = atom
        elif atom.name == "CZ":
            cz_atom = atom
        elif atom.name == "NE":
            ne_atom = atom
        elif atom.name == "NH1":
            nh1_atom = atom
        elif atom.name == "NH2":
            nh2_atom = atom
    h1_atom = protonate_sp2(cd_atom, ne_atom, cz_atom)
    h1_atom.name = "HE"
    h2_atom = protonate_direction(nh1_atom, ne_atom, cz_atom)
    h2_atom.name = "HN1"
    h3_atom = protonate_direction(nh1_atom, ne_atom, cd_atom)
    h3_atom.name = "HN2"
    h4_atom = protonate_direction(nh2_atom, ne_atom, cz_atom)
    h4_atom.name = "HN3"
    h5_atom = protonate_direction(nh2_atom, ne_atom, h1_atom)
    h5_atom.name = "HN4"
    return [h1_atom, h2_atom, h3_atom, h4_atom, h5_atom]


def add_his_hydrogen(residue):
    """Adds His hydrogen atoms to residues according to the 'old way'.

    Args:
        residue:  histidine residue to protonate
    """
    for atom in residue:
        if atom.name == "CG":
            cg_atom = atom
        elif atom.name == "ND1":
            nd_atom = atom
        elif atom.name == "CD2":
            cd_atom = atom
        elif atom.name == "CE1":
            ce_atom = atom
        elif atom.name == "NE2":
            ne_atom = atom
    hd_atom = protonate_sp2(cg_atom, nd_atom, ce_atom)
    hd_atom.name = "HND"
    he_atom = protonate_sp2(cd_atom, ne_atom, ce_atom)
    he_atom.name = "HNE"


def add_trp_hydrogen(residue):
    """Adds Trp hydrogen atoms to residues according to the 'old way'.

    Args:
        residue:  tryptophan residue to protonate
    """
    cd_atom = None
    ne_atom = None
    for atom in residue:
        if atom.name == "CD1":
            cd_atom = atom
        elif atom.name == "NE1":
            ne_atom = atom
        elif atom.name == "CE2":
            ce_atom = atom
    if (cd_atom is None) or (ne_atom is None) or (ce_atom is None):
        str_ = "Unable to find all atoms for {0:s} {1:s}".format(
            residue[0].res_name, residue[0].res_num)
        raise ValueError(str_)
    he_atom = protonate_sp2(cd_atom, ne_atom, ce_atom)
    he_atom.name = "HNE"


def add_amd_hydrogen(residue):
    """Adds Gln & Asn hydrogen atoms to residues according to the 'old way'.

    Args:
        residue:  glutamine or asparagine residue to protonate
    """
    c_atom = None
    o_atom = None
    n_atom = None
    for atom in residue:
        if ((atom.res_name == "GLN" and atom.name == "CD")
                or (atom.res_name == "ASN" and atom.name == "CG")):
            c_atom = atom
        elif ((atom.res_name == "GLN" and atom.name == "OE1")
              or (atom.res_name == "ASN" and atom.name == "OD1")):
            o_atom = atom
        elif ((atom.res_name == "GLN" and atom.name == "NE2")
              or (atom.res_name == "ASN" and atom.name == "ND2")):
            n_atom = atom
    if (c_atom is None) or (o_atom is None) or (n_atom is None):
        str_ = "Unable to find all atoms for {0:s} {1:s}".format(
            residue[0].res_name, residue[0].res_num)
        raise ValueError(str_)
    h1_atom = protonate_direction(n_atom, o_atom, c_atom)
    h1_atom.name = "HN1"
    h2_atom = protonate_average_direction(n_atom, c_atom, o_atom)
    h2_atom.name = "HN2"


def add_backbone_hydrogen(residue, o_atom, c_atom):
    """Adds hydrogen backbone atoms to residues according to the old way.

    dR is wrong for the N-terminus (i.e. first residue) but it doesn't affect
    anything at the moment. Could be improved, but works for now.

    Args:
        residue:  residue to protonate
        o_atom:  backbone oxygen atom
        c_atom:  backbone carbon atom
    Returns:
        [new backbone oxygen atom, new backbone carbon atom]
    """
    new_c_atom = None
    new_o_atom = None
    n_atom = None
    for atom in residue:
        if atom.name == "N":
            n_atom = atom
        if atom.name == "C":
            new_c_atom = atom
        if atom.name == "O":
            new_o_atom = atom
    if None in [c_atom, o_atom, n_atom]:
        return [new_o_atom, new_c_atom]
    if n_atom.res_name == "PRO":
        # PRO doesn't have an H-atom; do nothing
        pass
    else:
        h_atom = protonate_direction(n_atom, o_atom, c_atom)
        h_atom.name = "H"
    return [new_o_atom, new_c_atom]


def protonate_direction(x1_atom, x2_atom, x3_atom):
    """Protonates an atom, x1_atom, given a direction.

    New direction for x1_atom proton is (x2_atom -> x3_atom).

    Args:
        x1_atom:  atom to be protonated
        x2_atom:  atom for direction
        x3_atom:  other atom for direction
    Returns:
        new hydrogen atom
    """
    dx = (x3_atom.x - x2_atom.x)
    dy = (x3_atom.y - x2_atom.y)
    dz = (x3_atom.z - x2_atom.z)
    length = math.sqrt(dx*dx + dy*dy + dz*dz)
    x = x1_atom.x + dx/length
    y = x1_atom.y + dy/length
    z = x1_atom.z + dz/length
    h_atom = make_new_h(x1_atom, x, y, z)
    h_atom.name = "H"
    return h_atom


def protonate_average_direction(x1_atom, x2_atom, x3_atom):
    """Protonates an atom, x1_atom, given a direction.

    New direction for x1_atom is (x1_atom/x2_atom -> x3_atom).
    Note, this one uses the average of x1_atom & x2_atom (N & O) unlike
    the previous N - C = O

    Args:
        x1_atom:  atom to be protonated
        x2_atom:  atom for direction
        x3_atom:  other atom for direction
    Returns:
        new hydrogen atom
    """
    dx = (x3_atom.x + x1_atom.x)*0.5 - x2_atom.x
    dy = (x3_atom.y + x1_atom.y)*0.5 - x2_atom.y
    dz = (x3_atom.z + x1_atom.z)*0.5 - x2_atom.z
    length = math.sqrt(dx*dx + dy*dy + dz*dz)
    x = x1_atom.x + dx/length
    y = x1_atom.y + dy/length
    z = x1_atom.z + dz/length
    h_atom = make_new_h(x1_atom, x, y, z)
    h_atom.name = "H"
    return h_atom


def protonate_sp2(x1_atom, x2_atom, x3_atom):
    """Protonates a SP2 atom, given a list of atoms

    Args:
        x1_atom:  atom to set direction
        x2_atom:  atom to be protonated
        x3_atom:  other atom to set direction
    Returns:
        new hydrogen atom
    """
    dx = (x1_atom.x + x3_atom.x)*0.5 - x2_atom.x
    dy = (x1_atom.y + x3_atom.y)*0.5 - x2_atom.y
    dz = (x1_atom.z + x3_atom.z)*0.5 - x2_atom.z
    length = math.sqrt(dx*dx + dy*dy + dz*dz)
    x = x2_atom.x - dx/length
    y = x2_atom.y - dy/length
    z = x2_atom.z - dz/length
    h_atom = make_new_h(x2_atom, x, y, z)
    h_atom.name = "H"
    return h_atom


def make_new_h(atom, x, y, z):
    """Add a new hydrogen to an atom at the specified position.

    Args:
        atom:  atom to protonate
        x:  x position of hydrogen
        y:  y position of hydrogen
        z:  z position of hydrogen
    Returns:
        new hydrogen atom
    """
    new_h = Atom()
    new_h.set_property(
        numb=None, name='H{0:s}'.format(atom.name[1:]),
        res_name=atom.res_name, chain_id=atom.chain_id,
        res_num=atom.res_num, x=x, y=y, z=z, occ=None, beta=None)
    new_h.element = 'H'
    new_h.bonded_atoms = [atom]
    new_h.charge = 0
    new_h.steric_number = 0
    new_h.number_of_lone_pairs = 0
    new_h.number_of_protons_to_add = 0
    new_h.num_pi_elec_2_3_bonds = 0
    atom.bonded_atoms.append(new_h)
    atom.conformation_container.add_atom(new_h)
    return new_h


